/**
 * ActionManager: "static" listener to action and
 * item events, and redirect the event to a proper
 * function, the advantage is that the function can
 * be whatevernameyouwant(ActionEvent e) and it can
 * be initialized automatically to receive the event
 * transparently if it has the right name with the Bundle
 * class (or an extension).
 * @version 1.0
 * @author Alexandre David
 */

package adavid.awt;

import java.awt.Component;
import java.awt.event.*;
import adavid.reflect.ObjectMethod;
import java.util.Hashtable;
import java.lang.reflect.InvocationTargetException;
import adavid.reflect.ClassReflect;

/**
 * ActionManager
 * contains an ActionListener which redirects the calls
 * to specialized functions of the form
 * public void/whatever something(ActionEvent e)
 * only one redirection per component!
 */
public class ActionManager implements ActionListener,ItemListener {

  /**
   * this garantees that there is only one instance
   */
  public static final ActionManager listener=new ActionManager();

  private Hashtable handlerTable=new Hashtable();

  /**
   * the Class object to create the methods
   */
  public static final Class
      ACTIONEVENT=ClassReflect.getClass("java.awt.event.ActionEvent"),
      ITEMEVENT=ClassReflect.getClass("java.awt.event.ItemEvent");
  public static final Class[] ARG_ACTIONEVENT=new Class[] { ACTIONEVENT };
  public static final Class[] ARG_ITEMEVENT=new Class[] { ITEMEVENT };
  
  /**
   * add an action listener
   */
  public static void add(Object compo,String func,Object handler)
       throws NoSuchMethodException, SecurityException {

	 ObjectMethod method=new ObjectMethod(handler,func,ARG_ACTIONEVENT);
	 synchronized(listener.handlerTable) {
	   listener.handlerTable.put(compo,method);
	 }
  }
  
  /**
   * add an item listener
   */
  public static void addItem(Object compo,String func,Object handler)
       throws NoSuchMethodException, SecurityException {

	 ObjectMethod method=new ObjectMethod(handler,func,ARG_ITEMEVENT);
	 synchronized(listener.handlerTable) {
	   listener.handlerTable.put(compo,method);
	 }
  }

  /**
   * remove the listened object
   */
  public static void remove(Component compo) {
    synchronized(listener.handlerTable) {
      listener.handlerTable.remove(compo);
    }
  }

  /**
   * implements ActionListener
   * redirection
   */
  public void actionPerformed(ActionEvent event) {
    try {
      ObjectMethod method;
      synchronized(handlerTable) {
	method=(ObjectMethod)handlerTable.get(event.getSource());
      }
      //System.out.println("invoke "+method);
      method.invoke(new Object[] { event });
    }
    catch(InvocationTargetException ite) {
      System.err.println("actionPerformed("+event+"):");
      ite.getTargetException().printStackTrace();
    }
    catch(IllegalAccessException iae) {
      System.err.println("actionPerformed("+event+"):");
      iae.printStackTrace();
    }
  }

  /**
   * implements ItemListener
   * redirection
   */
  public void itemStateChanged(ItemEvent event) {
    try {
      ObjectMethod method;
      synchronized(handlerTable) {
	method=(ObjectMethod)handlerTable.get(event.getItemSelectable());
      }
      //System.out.println("invoke "+method);
      method.invoke(new Object[] { event });
    }
    catch(InvocationTargetException ite) {
      System.err.println("itemStateChanged("+event+"):");
      ite.getTargetException().printStackTrace();
    }
    catch(IllegalAccessException iae) {
      System.err.println("itemStateChanged("+event+"):");
      iae.printStackTrace();
    }
  }

}


