/**
 * DialogFactory: a factory for simple dialog windows
 * @version 1.0.4
 * @author Alexandre David
 */

package adavid.awt;

/**
 * version history:
 * 1.0: first version
 * 1.0.1: showMessage: added a null pointer detection
 * 1.0.2: public constructor
 * 1.0.3 991011: trim() for the input TextField
 * 1.0.4 991114: constants changed
 */

import java.awt.*;
import java.awt.event.*;
import adavid.util.Bundle;

/**
 * DialogFactory
 * simple dialog windows
 * swing free
 */
public class DialogFactory extends Dialog implements ActionListener {

  public static int border=12;
  public static final int MAX_WIDTH=600,MAX_HEIGHT=500;

  /**
   * the button that was pressed in the dialog
   */
  private Object pressedButton=null;

  /**
   * the find button
   */
  private NiceButton findButton=null;

  /**
   * the find text field
   */
  private TextField input;

  /**
   * the text area: the message
   */
  private TextArea text;

  /**
   * just the constructor: not public because the user is not
   * supposed to call it ; changed now: useful to have the
   * close predefined! sorry for that.
   */
  public DialogFactory(Frame parent,String title,boolean mode) {
    super(parent,title,mode);
    enableEvents(AWTEvent.WINDOW_EVENT_MASK);
  }

  /**
   * open a simple message dialog box
   */
  public static void showMsg(Frame parent,String title,String msgtxt,String closemsg) {
    DialogFactory dialog=new DialogFactory(parent,title,false);
    Graphics g=parent.getGraphics();
    NiceLabel msg=new NiceLabel(g,msgtxt,Label.CENTER);
    NiceButton close=new NiceButton(g,closemsg);
    Panel panel=new Panel(null);
    close.addActionListener(dialog);

    int h=close.prefHeight();
    int wm=close.prefWidth();
    int w=Math.max(msg.prefWidth(),wm);
    int freew=w-wm;

    dialog.add(panel);
    panel.add(msg);
    panel.add(close);
    
    msg.setBounds(border,border,w,h);
    close.setBounds(border+freew/2,2*border+h,wm,h);
    panel.setSize(w+2*border,2*h+3*border);
    dialog.pack();
    dialog.setResizable(false);
    Bundle.show(parent,dialog);
  }

  /**
   * open a yes/no dialog box type and return true if the yes
   * was pressed. Closing the window=no
   */
  public static boolean confirmYN(Frame parent,String title,
				  String question,String yesName,String noName) {
    DialogFactory dialog=new DialogFactory(parent,title,true);
    Graphics g=parent.getGraphics();
    NiceLabel msg=new NiceLabel(g,question,Label.CENTER);
    NiceButton yes=new NiceButton(g,yesName),no=new NiceButton(g,noName);
    Panel panel=new Panel(null);

    yes.addActionListener(dialog);
    no.addActionListener(dialog);

    int h=yes.prefHeight();
    int w2=yes.prefWidth();
    int w3=no.prefWidth();
    int w=Math.max(msg.prefWidth(),w2+w3);
    int freew=w-w2-w3;

    dialog.add(panel);
    panel.add(msg);
    panel.add(yes);
    panel.add(no);
    panel.setBackground(parent.getBackground().darker());

    msg.setBounds(border,border,w,h);
    yes.setBounds(border+freew/3,2*border+h,w2,h);
    no.setBounds(border+2*freew/3+w2,2*border+h,w3,h);
    panel.setSize(w+2*border,2*h+3*border);
    dialog.pack();
    //Under JDK 1.1.7, this causes a core dump
    dialog.setResizable(false);
    Bundle.show(parent,dialog);
    //Under Kaffe, it does not return here!
    boolean answer=yes.equals(dialog.pressedButton);
    dialog.dispose();
    return answer;
  }

  /**
   * show a dialog alert message window
   */
  public static void showAlert(Frame parent,String title,String message,String closeLabel) {
    showMessage(parent,title,message,closeLabel,null);
  }

  /**
   * show a message window with a find query text field
   */
  public static void showMessage(Frame parent,String title,String message,String closeLabel,String query) {
    DialogFactory dialog=new DialogFactory(parent,title,false);
    Graphics g=parent.getGraphics();
    NiceButton closeButton=new NiceButton(g,closeLabel);
    if (message==null) message="null";
    int n=1,index=0,len=message.length(),ncol=0,col=0;
    while (n<40 && index<len)
      if (message.charAt(index++)=='\n') { n++; ncol=Math.max(col,ncol); col=0; }
    else
      col++;
    ncol=Math.max(col,ncol+2);
    if (ncol>80)
      ncol=80;
    else if (ncol<40)
      ncol=40;
    dialog.text=new TextArea(message,(n<5)?5:n,ncol);
    Panel panel=new Panel(new BorderLayout());
    Panel south=new Panel(new FlowLayout(FlowLayout.CENTER,border/4,border/4));

    dialog.text.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
    closeButton.addActionListener(dialog);
    dialog.add(panel);
    panel.add(dialog.text,BorderLayout.CENTER);
    panel.add(south,BorderLayout.SOUTH);
    south.add(closeButton);

    if (query!=null) {
      Panel north=new Panel(new BorderLayout(2,2));
      north.add(new NiceLabel(g,query+":"),BorderLayout.WEST);
      dialog.input=new TextField();
      AutoFocuser.add(dialog.input);
      north.add(dialog.input,BorderLayout.CENTER);
      dialog.findButton=new NiceButton(g,query);
      north.add(dialog.findButton,BorderLayout.EAST);
      panel.add(north,BorderLayout.NORTH);
      dialog.findButton.addActionListener(dialog);
      dialog.input.addActionListener(dialog);
      dialog.text.setBackground(Color.white);
    }
    else {
      dialog.text.setBackground(dialog.getBackground());
      dialog.text.setEditable(false);
    }

    dialog.pack();
    Dimension dim=dialog.getSize();
    if (dim.width>MAX_WIDTH)
      dim.width=MAX_WIDTH;
    if (dim.height>MAX_HEIGHT)
      dim.height=MAX_HEIGHT;
    dialog.setSize(dim);
    Bundle.show(parent,dialog);
  }

  /**
   * process window events: just close the window when asked
   */
  protected void processWindowEvent(WindowEvent e) {
    super.processWindowEvent(e);
    if (e.getID()==WindowEvent.WINDOW_CLOSING) {
      setVisible(false);
      if (!isModal())
	dispose();
    }
  }

  /**
   * ActionListener
   */
  public void actionPerformed(ActionEvent e) {
    Object source=e.getSource();
    if (source.equals(findButton) || source.equals(input)) {
      String findText=input.getText().trim();
      if (findText.length()>0) {
	String content=text.getText();
	int pos=content.indexOf(findText,text.getCaretPosition());
	if (pos<0 && text.getCaretPosition()>0)
	  pos=content.indexOf(findText);
	if (pos>=0)
	  text.select(pos,pos+findText.length());
      }
    }
    else {
      pressedButton=e.getSource();
      setVisible(false);
      if (!isModal())
	dispose();
    }
  }

    /**
     * @return a option dialog styled
     * @param parent the parent frame
     * @param color the color of the border
     * @param title the title of the dialog
     * @param subtitle the title of the borderedpanel
     * @param list the list of components
     * NOTE: the parent must have a valid graphics object!
     */
    public static DialogFactory getDialog(Frame parent,Color color,
					  String title,String subtitle,
					  Component[] list) {
	Graphics g=parent.getGraphics();
	DialogFactory dialog=new DialogFactory(parent,title,false);
	dialog.setLayout(new BorderLayout());
	dialog.add(BorderedPanel.getPanel(g,color,subtitle,list),BorderLayout.CENTER);
	dialog.pack();
	dialog.setResizable(false);
	return dialog;
    }

}

