/**
 * FLabel: focusable label, may get focus,
 * useful in trees
 * @version 1.0
 * @author Alexandre David
 */

package adavid.awt;

import java.awt.Label;
import java.awt.Graphics;
import java.awt.Color;
import java.awt.event.FocusEvent;
import java.awt.Dimension;
import java.awt.AWTEvent;
import java.awt.Font;
import java.awt.FontMetrics;

/**
 * FLabel : focusable label ; may look like a NiceLabel
 */
public class FLabel extends Label {

  /**
   * boolean to manage the focus
   */
  boolean hasFocus=false;

  /**
   * remember the preferred size if the label
   * was constructed with a valid graphics object
   */
  int w=0,h=0,borderw=0,borderh=0;
  Graphics g;

  /**
   * we use directly the focus event mask to manage the focus,
   * it is simple and efficient
   *****
   * we overrides all the constructors of Label to
   * provide a flexible component
   */
  public FLabel() {
    this(null,LEFT);
  }

  public FLabel(String text) {
    this(text,LEFT);
  }

  public FLabel(String text,int pos) {
    super(text,pos);
    enableEvents(AWTEvent.FOCUS_EVENT_MASK);
  }

  /**
   * constructors to use the NiceLabel style
   */
  public FLabel(Graphics g) {
    this(g,null,LEFT,8,4,true);
  }
  public FLabel(Graphics g,String text) {
    this(g,text,LEFT,8,4,true);
  }
  public FLabel(Graphics g,String text,int align) {
    this(g,text,align,8,4,true);
  }
  public FLabel(Graphics g,String text,int align,int borderw,int borderh) {
    this(g,text,align,borderw,borderh,true);
  }
  public FLabel(Graphics g,String name,int align,int borderw,int borderh,boolean bold) {
    super(name,align);
    enableEvents(AWTEvent.FOCUS_EVENT_MASK);
    this.borderw=borderw;
    this.borderh=borderh;
    if (g!=null) {
      this.g=g;
      Font font=g.getFont();
      if (bold)
	setFont(font=new Font(font.getName(),Font.BOLD,font.getSize()));
      FontMetrics fm=g.getFontMetrics(font);
      setSize(w=fm.stringWidth(name)+borderw,
	      h=fm.getHeight()+borderh);
    }
  }

  /**
   * overrides setText to update w and h
   */
  public synchronized void setText(String text) {
    super.setText(text);
    Font font=getFont();
    if (g!=null && font!=null) {
      FontMetrics fm=g.getFontMetrics(font);
      w=fm.stringWidth(text)+borderw;
      h=fm.getHeight()+borderh;
    }
  }

  /**
   * overrides setFont in order to update h and w
   */
  public void setFont(Font font) {
    super.setFont(font);
    if (g!=null && font!=null) {
      FontMetrics fm=g.getFontMetrics(font);
      w=fm.stringWidth(getText())+borderw;
      h=fm.getHeight()+borderh;
    }
  } 

  /**
   * overrides the method in Component to enable the focus
   * gained by tab
   */
  public boolean isFocusTraversable() {
    return true;
  }

  /**
   * this is to update the focus boolean since a removed component
   * will not be notified of the changing of the focus
   */
  public void removeNotify() {
    hasFocus=false;
    super.removeNotify();
  }

  /**
   * we process the focus event here, basically by
   * updating the focus boolean and repainting the component
   */
  protected void processFocusEvent(FocusEvent e) {
    hasFocus=(e.getID()==FocusEvent.FOCUS_GAINED);
    Graphics g=getGraphics();
    if (g!=null) {
      Dimension dim=getSize();
      Color color=g.getColor();
      g.setColor(hasFocus?Color.black:getBackground());
      g.drawRect(0,0,dim.width-1,dim.height-1);
      g.setColor(color);
    }
    super.processFocusEvent(e);//in case someone overrides this component
  }

  /**
   * paint the component: the LabelPeer has been called before,
   * the only interest here is to paint the focus border
   */
  public void paint(Graphics g) {
    //super.paint(g);//we do not need this, the super function is empty
    if (hasFocus) {
      Dimension dim=getSize();
      Color color=g.getColor();
      g.setColor(Color.black);
      g.drawRect(0,0,dim.width-1,dim.height-1);
      g.setColor(color);
    }
  }

  /**
   * returns the preferred size
   */
  public Dimension getPreferredSize() {
    if (w!=0 && h!=0)
      return new Dimension(w,h);
    else
      return super.getPreferredSize();
  }

  /**
   * returns preferred dimensions (w,h)
   */
  public int prefWidth() {
    if (w!=0)
      return w;
    else
      return super.getPreferredSize().width;
  }
  public int prefHeight() {
    if (h!=0)
      return h;
    else
      return super.getPreferredSize().height;
  }

}
  
