/**
 * ImageTNodeRenderer: implementation of a TNodeRenderer
 * with predefined icons. Just classical icons which look
 * like the file icons that anyone is used to see.
 * @version 1.0
 * @author Alexandre David
 */

package adavid.awt;

import java.awt.*;
import adavid.util.Bundle;

/**
 * default tree renderer, with image
 */
public class ImageTNodeRenderer implements TNodeRenderer {

  /**
   * the shared instance of this class ; shared since its
   * only purpose is to draw an image ; I adopt here a weak
   * programming model, namely the icons are assumed not to
   * be modified by a process while the instance is rendering
   * them, or updating the size information. The icons should be
   * set before. If this is not the case, it is easy to extend
   * this class and use a "synchronized(..)" on the instance, which
   * can serve as a lock, and use the methods of this class
   */
  public static final ImageTNodeRenderer INSTANCE=new ImageTNodeRenderer();

  /**
   * the modes
   */
  public static int FILES_AND_DIR=0,FILES_ONLY=1,DIR_ONLY=2;

  /**
   * constants to draw the triangle
   */
  public static final int SIZE=8;
  public static final int[] pictx={SIZE,SIZE/2,0};
  public static final int[] picty={0,SIZE,0};

  int mode; //store the renderer mode

  /**
   * dimension of the rendered image
   */
  protected static Dimension dim=new Dimension();

  /**
   * the path of the default icons, so that Bundle finds them
   */
  public static String path$icons="adavid/icons/";

  /**
   * the icons to be drawn
   */
  public static Image TreeLeaf,TreeLeafSelected;
  public static Image TreeNode,TreeSelected;
  public static Image TreeOpen,TreeOpenSelected;

  /**
   * basically, the initialization of the icons
   */
  static {
    Bundle bundle=new Bundle(INSTANCE.getClass());
    bundle.init(INSTANCE,null);
    updateDim();
  }

  /**
   * protected in order not to be called directly and
   * still allow extension
   */
  protected ImageTNodeRenderer() {
    this(FILES_AND_DIR);
  }

  /**
   * and a configurable renderer
   */
  public ImageTNodeRenderer(int mode) {
    this.mode=mode;
  }

  /**
   * update the size information of the icons
   */
  public static void updateDim() {
    synchronized(dim) {
      dim.width=0;
      dim.height=0;
      lookDim(TreeLeaf,dim);
      lookDim(TreeLeafSelected,dim);
      lookDim(TreeNode,dim);
      lookDim(TreeSelected,dim);
      lookDim(TreeOpen,dim);
      lookDim(TreeOpenSelected,dim);
      dim.width+=SIZE+2;
      if (dim.height<SIZE+1)
	dim.height=SIZE+1;
    }
  }

  /**
   * update the dimension if the needed size is
   * bigger than this dimension
   */
  public static void lookDim(Image img,Dimension checkDim) {
    if (img!=null && checkDim!=null) {
      int h=img.getHeight(null);
      int w=img.getWidth(null);
      if (h<0 || w<0) { //image possibly not ready
	if (!Bundle.waitImage(img))//then wait it
	  return;//this means error
	h=img.getHeight(null);
	w=img.getWidth(null);
      }
      if (h>checkDim.height)
	checkDim.height=h;
      if (w>checkDim.width)
	checkDim.width=w;
    }
  }   

  /**
   * we render the node here: draw the icon and the triangle
   * note that the TNode chooses the color of the triangle
   */
  public void paint(TNode c,Graphics g,boolean isExpanded,boolean isSelected,boolean isLeaf) {
    Image icon;
    if ((isLeaf && mode==FILES_AND_DIR) || mode==FILES_ONLY) {
      icon=(isSelected)?TreeLeafSelected:TreeLeaf;
      if (mode==FILES_ONLY && !isLeaf) {
	if (isExpanded)
	  g.fillPolygon(pictx,picty,3);
	else
	  g.fillPolygon(picty,pictx,3);
      }
    }
    else if (isExpanded && (!isLeaf || mode!=DIR_ONLY)) {
      if (!isLeaf)
	g.fillPolygon(pictx,picty,3);
      icon=(isSelected)?TreeOpenSelected:TreeOpen;
    }
    else {
      if (!isLeaf)
	g.fillPolygon(picty,pictx,3);
      icon=(isSelected)?TreeSelected:TreeNode;
    }
    if (icon!=null)
      g.drawImage(icon,SIZE+1,0,icon.getWidth(null),icon.getHeight(null),c);
  }

  /**
   * the dimension taken by the rendered image and triangle ;
   * updateDim() is assumed to be called before ; has to be done
   * explicitely if one changes one icon
   */
  public int getHeight() { return dim.height; }
  public int getWidth()  { return dim.width;  }

}
