/**
 * TNode: the tree node object
 * @version 1.0.1
 * @author Alexandre David
 */

/**
 * version history
 * 1.0: first version
 * 1.0.1: 990905 changed the step
 */

package adavid.awt;

/**
 * no Vector used, local management, to reduce memory usage, gain in
 * speed, and reduce the casts
 */
import java.awt.*;
import java.awt.event.*;

/**
 * TNode : Tree Node
 * I create my own tree component because Java copies the nodes
 * without telling me when, and it copies the content of the nodes
 * as well, so when my other components refer to these nodes, well,
 * they are not valid because they have been replaced by new ones!
 * Imagine the time I've lost to understand where the problem came
 * from, and then to implement my own tree!!!
 * This is my second version of tree: it is 100% swing free, only
 * pure java awt, which is much much faster than swing.
 *
 * A node listens to its own events to expand/collapse or send an event
 * to a listener ; to use the serialization the user has to be sure that
 * the content object, the TListener objects and the TNodeRenderer objects
 * are serializable as well!!!!!
 */
public class TNode extends Container implements MouseListener,KeyListener {

  /**
   * default color settings: the background when selected
   * and the default color for the renderer
   */
  public static final Color DEFAULT_SELECTED_COLOR=new Color(204,204,255);
  public static final Color DEFAULT_COLOR=Color.blue;

  /**
   * step used to enlarge the array of leaves/children when adding a node
   * useful when dealing with big trees to reduce memory allocation
   */
  public static int step=32;

  /**
   * current colors, set to the default values
   */
  Color colorSelected=DEFAULT_SELECTED_COLOR,color=DEFAULT_COLOR;

  /**
   * the parent node, to which this node is attached, and the selected
   * node ; the selection node is used only by the root node (parent==null)
   */
  TNode parent,selection;

  /**
   * the children/leaves of this node ; a check is done on step, since
   * it can be modified without control ; nbNodes is the actual number
   * of nodes used in this array.
   */
  TNode[] children=new TNode[(step>0)?step:32];
  int nbNodes=0;

  /**
   * the content of the node ; any object
   */
  Object content;

  /**
   * a flag to indicate that the content changes or
   * not: to update the label(?) or not
   */
  boolean mutable=true;

  /**
   * the representation of this node ; any Component
   */
  Component item;

  /**
   * the list of listeners and the actual number of listeners in the array
   */
  TListener[] listeners;
  int nbListeners;

  /**
   * customize double click->open/close
   */
  boolean openOnDoubleClick=true;

  /**
   * the dimension of this node ; this includes the leaves as well
   */
  transient Dimension nodeSize=new Dimension();

  /**
   * booleans representing the state of this node:
   *  valid: the size is not valid and needs to be computed
   *  expanded: node expanded
   *  selected: node selected
   */
  transient boolean valid=false,expanded=false,selected=false;

  /**
   * the (optional) scroll pane in which the node can be ; used to send repaint
   * events when expanding/collapsing the node
   */
  ScrollPane scroll;

  /**
   * transient graphical information
   */
  transient int shift,lineHeight;

  /**
   * the renderer of a node ; should be serializable if one wants to serialize
   * this node
   */
  TNodeRenderer renderer;

  /**
   * PopupMenu to be opened
   */
  PopupMenu popup;

  /**
   * contructor : take the content ; if it is a component, it will display
   * itself, otherwise a label will be created with its toString()
   * the DefaultTNodeRenderer is used
   */
  public TNode(Object object) {
    this(object,(TNodeRenderer)null);
  }

  /**
   * constructor : same as previously with a customized renderer
   */
  public TNode(Object object,TNodeRenderer renderer) {
    this(object,renderer,
	 (object instanceof Component || object==null)?
	 (Component)object:
	 new FLabel(object.toString()));
  }

  /**
   * another constructor
   */
  public TNode(Object object,Component item) {
    this(object,null,(item!=null)?item:
	 ((object instanceof Component || object==null)?(Component)object:
	  new FLabel(object.toString())));
  }

  /**
   * real constructor : takes the content (object), the renderer
   * and a component to display the content ; the component should
   * be able to update itself when its background changes.
   */
  public TNode(Object object,TNodeRenderer renderer,Component item) {
    this.parent=null;
    setRenderer(renderer);
    setLayout(null);
    content=object;
    if (item==null && object!=null)
      item=new FLabel(object.toString());
    this.item=item;
    if (item!=null) {
      addMouseListener(this);
      item.addMouseListener(this);
      add(item);
      if (item instanceof Label)
	item.addKeyListener(this);
    }
  }

  /**
   * returns the content of the node ; the content is immutable
   */
  public Object getContent() {
    return content;
  }

  /**
   * sets the mutable flag: mutable is inherited
   */
  public void setMutable(boolean flag) {
    mutable=flag;
  }

  /**
   * returns true if the node is mutable
   */
  public boolean isMutable() {
    return mutable;
  }

  /**
   * returns the component of the TNode
   */
  public Component getComponent() {
    return item;
  }

  /**
   * customize the double click->open/close
   */
  public void setOpenOnDoubleClick(boolean flag) {
    openOnDoubleClick=flag;
  }
  public boolean getOpenOnDoubleClick() {
    return openOnDoubleClick;
  }

  /**
   * add/set the popup menu
   */
  public synchronized void add(PopupMenu pmenu) {
    if (pmenu!=null) {
      remove(popup);
      popup=pmenu;
      super.add(pmenu);
    }
  }

  /**
   * remove the popup menu
   */
  public synchronized void remove(MenuComponent mc) {
    if (mc!=null) {
      super.remove(mc);
      popup=null;
    }
  }

  /**
   * change the font : resend this to the component which represents the content
   * of this node
   */
  public void setFont(Font font) {
    super.setFont(font);
    if (item!=null)
      item.setFont(font);
  }

  /**
   * change the renderer ; test if it is not null
   */
  public void setRenderer(TNodeRenderer renderer) {
    this.renderer=(renderer==null)?DefaultTNodeRenderer.INSTANCE:renderer;
  }

  /**
   * returns the color used for background selection
   */
  public Color getColorSelected() {
    return colorSelected;
  }

  /**
   * returns the color used for the renderer ; the Graphics
   * object has a g.setColor(..) just before the call to the
   * renderer
   */
  public Color getColor() {
    return color;
  }

  /**
   * change the color for the renderer
   */
  public void setColor(Color color) {
    if (color!=null)
      this.color=color;
  }

  /**
   * change the color for the background selection
   */
  public void setColorSelected(Color color) {
    if (color!=null)
      colorSelected=color;
  }

  /**
   * set the scroll pane (useful for the root only)
   */
  public void setScrollPane(ScrollPane pane) {
    scroll=pane;
  }

  /**
   * delete this node ; works only if this is not
   * the root ; delegate the call to its parent
   */
  public void delete() {
    if (parent!=null)
      parent.removeNode(this);
  }

  /**
   * returns the parent node
   */
  public TNode getParentNode() {
    return parent;
  }

  /**
   * returns the expansion state
   */
  public boolean isExpanded() {
    return expanded;
  }

  /**
   * returns all the sub nodes
   */
  public TNode[] getSubNodes() {
    TNode[] result=new TNode[nbNodes];
    System.arraycopy(children,0,result,0,nbNodes);
    return result;
  }

  /**
   * returns one sub node
   * null if invalid
   */
  public TNode getSubNode(int index) {
    if (index>=0 && index<nbNodes)
      return children[index];
    else
      return null;
  }

  /**
   * return the string of the ?label
   */
  public String getStringLabel() {
    if (item !=null && item instanceof Label)
      return ((Label)item).getText();
    else
      return (content==null)?null:content.toString();
  }

  /**
   * returns the content of all sub nodes
   */
  public Object[] getSubContents() {
    Object[] result=new Object[nbNodes];
    for(int i=0;i<nbNodes;i++)
      result[i]=children[i].content;
    return result;
  }

  /**
   * add a new node to this tree ; bad things may happen
   * if one adds twice the same node
   */
  public void addNode(TNode node) {
    if (node!=null) {
      if (nbNodes>=children.length) {
	TNode[] old=children;
	if (step<=0)
	  step=32;
	children=new TNode[nbNodes+step];
	System.arraycopy(old,0,children,0,old.length);
      }
      children[nbNodes++]=node;
      //the added node inherits the properties of its
      //parent ; to customize this node, one has to call
      //the appropriate functions after addNode(..)
      node.parent=this;
      if (node.listeners==null) {
	node.listeners=listeners;
	node.nbListeners=nbListeners;
      }
      if (node.popup==null)
	node.popup=popup;
      node.colorSelected=colorSelected;
      node.color=color;
      node.renderer=renderer;
      node.openOnDoubleClick=openOnDoubleClick;
      node.mutable=mutable;
      if (item!=null && node.item!=null)
	node.item.setCursor(item.getCursor());
      if (expanded) {
	add(node);
	invalidateNode(node);
      }
    }
  }

  /**
   * invalidate the node given in arguments
   * if one node is not valid, then all its parents
   * are not valid, so we do not need to propagate when we
   * find one !valid, thus the test in the loop
   */
  public static void invalidateNode(TNode node) {
    while (node!=null && node.valid) {
      node.valid=false;
      node=node.parent;
    }
  }

  /**
   * invalidates itself and revalidates itself ; should be called
   * after some addNode in order to update the layout
   */
  public void invalidateNode() {
    invalidateNode(this);
    getRootNode().getPreferredSize();
  }

  /**
   * removes a node ; local management of memory as in Vector
   */
  public void removeNode(TNode node) {
    int i=0;
    while(i<nbNodes && children[i]!=node) i++;
    if (i<nbNodes) {
      nbNodes--;
      if (i<nbNodes)
	System.arraycopy(children,i+1,children,i,nbNodes-i);
      remove(node);
      if (node.item!=null) {
	node.item.removeMouseListener(this);
	if (node.item instanceof Label)
	  node.item.removeKeyListener(this);
      }
      invalidateNode(this);
    }
  }

  /**
   * remove all nodes
   */
  public void removeAllNodes() {
    for(int i=0;i<nbNodes;i++) {
      Component itemi=children[i].item;
      remove(children[i]);
      if (itemi!=null) {
	itemi.removeMouseListener(this);
	if (itemi instanceof Label)
	  itemi.removeKeyListener(this);
      }
    }
    nbNodes=0;
    invalidateNode(this);
  }

  /**
   * expand this node ; and all the path to this node to make it visible
   */
  public void expand() {
    if (!expanded) {
      expanded=true;
      for(int i=0;i<nbNodes;i++)
	add(children[i]);
    }
    valid=false;
    if (parent!=null)
      parent.expand();
  }

  /**
   * expand all children and invalidate
   */
  public void expandAll() {
    expandAllChildren();
    if (parent!=null)
      parent.expand();
    TNode root=getRootNode();
    root.getPreferredSize();
    if (root.scroll!=null)
      root.scroll.repaint();
    else if (root.getParent()!=null)
      root.getParent().repaint();
   }

  /**
   * expand all children
   */
  void expandAllChildren() {
    expanded=true;
    for(int i=0;i<nbNodes;i++) {
      children[i].expandAll();
      add(children[i]);
    }
    valid=false;
  }

  /**
   * collapse this node
   */
  public void collapse() {
    if (expanded) {
      expanded=false;
      int n=getComponentCount();
      while(n-->=0 && getComponent(n)!=item)
	remove(n);
      invalidateNode(this);
    }
  }

  /**
   * string representation of this node
   */
  public String toString() {
    return new StringBuffer(getClass().getName()).append(':').append(content).toString();
  }

  /**
   * returns the preferred size of this node ; this does the computation
   * of the size when not valid and does the layout as well
   */
  public Dimension getPreferredSize() {
    if (!valid) {
      int h=0,w=0;
      int rw=0,first=0;

      //the content first : label or general component
      if (item!=null) {
	if (mutable && item instanceof Label && content!=null)
	  ((Label)item).setText(content.toString());
	Dimension dim=item.getPreferredSize();
	int rh=renderer.getHeight();
	w=dim.width+(rw=renderer.getWidth());
	item.setBounds(rw,0,dim.width,h=dim.height);
	if (rh>h)
	  h=rh;
	lineHeight=h;
	shift=h/2-rh/2;
	first=1;
      }

      //all the sub-component
      Component[] components=getComponents();
      int n=components.length;
      for(int i=first;i<n;i++) {
	Dimension dim=components[i].getPreferredSize();
	components[i].setLocation(rw,h);
	w=Math.max(w,dim.width+rw);
	h+=dim.height;
      }
      
      nodeSize.width=w;
      nodeSize.height=h;
      valid=true;
      setSize(new Dimension(nodeSize));
      
      if (parent==null && scroll!=null) {
	  scroll.invalidate();
	  scroll.doLayout();
      }
    }

    return new Dimension(nodeSize);
  }

  /**
   * the paint method: delegate the call to the renderer
   * the component is called by the super method
   */
  public void paint(Graphics g) {
    if (item!=null) {
      g.translate(0,shift);
      g.setColor(getBackground());
      g.fillRect(0,0,renderer.getWidth(),renderer.getHeight());
      g.setColor(color);
      renderer.paint(this,g,expanded,selected,nbNodes==0);
      g.translate(0,-shift);
    }
    super.paint(g);
  } 

  /**
   * this is to clean the component before repainting it
   */
  public void repaint(long tm, int x, int y, int width, int height) {
    Graphics g=getGraphics();
    if (g!=null) {
      g.setColor(getBackground());
      g.fillRect(x,y,width,height);
    }
    //System.out.println(x+","+y+","+width+","+height);
    super.repaint(tm,x,y,width,height);
  }

  /**
   * add a TNode listener
   */
  public void addTListener(TListener listener) {
    if (listeners==null) {
      listeners=new TListener[5];
      nbListeners=0;
    }
    else if (nbListeners>=listeners.length) {
      TListener[] old=listeners;
      listeners=new TListener[nbListeners+5];
      System.arraycopy(old,0,listeners,0,old.length);
    }
    listeners[nbListeners++]=listener;
  }

  /**
   * remove a TNode listener
   */
  public void removeTListener(TListener listener) {
    if (listener!=null) {
      int i=0;
      while(i<nbListeners && listeners[i]!=listener) i++;
      if (i<nbListeners) {
	nbListeners--;
	if (i<nbListeners)
	  System.arraycopy(listeners,i+1,listeners,i,nbListeners-1);
      }
    }
  }

  /**
   * unselects itself : repaints itself ; the component which represents this
   * node should be able to repaint itself on its own after the setBackground call
   */
  public void unselect() {
    TNode node=getRootNode();
    TNode current=node.selection;
    if (current==this) {
      current.selected=false;
      if (current.item!=null)
	current.item.setBackground(getBackground());
      current.repaint(0,0,0,current.renderer.getWidth(),current.lineHeight);
      current.validate();
    }
    node.selection=null;
  }
  
  /**
   * returns the root node of this node
   */
  public TNode getRootNode() {
    TNode node=this;
    while(node.parent!=null)
      node=node.parent;
    return node;
  }

  /**
   * selects this node ; unselect the selected one and selects itself
   * see unselect for the repaint
   */
  public void select() {
    TNode node=getRootNode();
    TNode current=node.selection;
    if (current!=null && current!=this) {
      current.selected=false;
      if (current.item!=null)
	current.item.setBackground(getBackground());
      current.repaint(0,0,0,current.renderer.getWidth(),current.lineHeight);
      current.validate();
    }
    node.selection=this;
    selected=true;
    if (item!=null)
      item.setBackground(colorSelected);
    repaint(0,0,0,renderer.getWidth(),lineHeight);
    validate();
  }

  /**
   * toggle between collapse/expand and repaint
   */
  public void toggle() {
    if (expanded)
      collapse();
    else
      expand();
    TNode root=getRootNode();
    root.getPreferredSize();
    if (root.scroll!=null)
      root.scroll.repaint();
    else if (root.getParent()!=null)
      root.getParent().repaint();
  }

  /**
   * mouse listener ; delegate the call to the listeners
   * after processing (expand/collapse)
   */
  public void mouseClicked(MouseEvent e) {
    if (e.getSource()!=this) {
      int count=e.getClickCount();
      if (count==1) {
	select();
	if (item!=null)
	  item.requestFocus();
	for(int i=0;i<nbListeners;i++)
	  listeners[i].nodeClicked(this);
      }
      else if (count==2) {
	if (openOnDoubleClick)
	  toggle();
	//item.requestFocus();//has already the focus
	for(int i=0;i<nbListeners;i++)
	  listeners[i].nodeDoubleClicked(this);
      }
    }
    else if (item!=null && e.getY()<=lineHeight && e.getX()<=renderer.getWidth()) {
      select();
      item.requestFocus();
      toggle();
    }
  }

  public void mousePressed(MouseEvent e) {
    if (popup!=null && e.isPopupTrigger())
      popup.show(this,e.getX(),e.getY());
  }

  /**
   * same as before ; we call the nodeClicked method as if it were
   * a click when one presses the space bar ; the return key just toggles
   * the node
   */
  public void keyTyped(KeyEvent e) {
    char c=e.getKeyChar();
    if (c==' ') {
      select();
      //item.requestFocus();//must have the focus to get the key event
      for (int i=0;i<nbListeners;i++)
	listeners[i].nodeClicked(this);
    }
    else if ((c=='\n' || c=='\r') && nbNodes>0)
      toggle();
    else if (c=='*')
      expandAll();
  }

  /**
   * unused methods of listeners
   * I prefer to use this method instead of using the
   * "empty" listener provided by Sun, which we should
   * extend, imagine, you download two classes just for
   * a simple listener, and you have to pass arguments
   * to have access to what you want. This is much simpler.
   */
  public void mouseReleased(MouseEvent e) {}
  public void mouseEntered(MouseEvent e) {}
  public void mouseExited(MouseEvent e) {}
  public void keyPressed(KeyEvent e) {}
  public void keyReleased(KeyEvent e) {}
}

