/**
 * ClassReflect
 * This gives more information on classes than Class
 * and is able to filter information.
 * @version 1.0.1
 * @author Alexandre David
 */

/**
 * version history
 * 1.0: first version
 * 1.0.1: 990818 use copyInto for all Vector -> array
 */

package adavid.reflect;

import adavid.util.*;
import java.lang.reflect.*;
import java.util.*;
import java.awt.Color;

/**
 * useful functions to perform operations on Class
 * without any instance
 */
public class ClassReflect {

  /**
   * flags to control the behaviour of the functions
   */
  static boolean shorterNames=true,shorterTypes=true,doSort=true,sortModifiers=false;

  /**
   * classes used for tests: better to have then once here,
   * and they are usable by other classes as well
   */
  public static final Class
      STRING=getClass("java.lang.String"),
      CHAR=Character.TYPE,
      CHARACTER=getClass("java.lang.Character"),
      STRINGBUFFER=getClass("java.lang.StringBuffer"),
      COLOR=getClass("java.awt.Color");
  
  /**
   * filter mode definition
   */
  public static final int PACKAGE=Modifier.PUBLIC | Modifier.PROTECTED | Modifier.PRIVATE,ALLACCESS=PACKAGE << 1;
  public static final int YES=1,NO=2,ANY=0;

  /**
   * flags to control the filter
   */
  static int accessFlag=Modifier.PUBLIC,finalFlag=ANY,staticFlag=ANY,abstractFlag=ANY,nativeFlag=ANY,synchronizedFlag=ANY,transientFlag=ANY,volatileFlag=ANY;

  /**
   * the color table
   */
  static Hashtable colorTable=new Hashtable();

  static {
      try {
	Field[] field=COLOR.getFields();
	for(int i=0;i<field.length;i++)
	  if (field[i].getType().getName().equals("java.awt.Color")) {
	    try {colorTable.put(field[i].get(null),field[i].getName());}
	    catch(Exception whatever){}
	  }
      }
      catch(Exception whatever){}
  }

  /**
   * used to initialize the Class static members
   */
  public static Class getClass(String name) {
    try {
      return Class.forName(name);
    }
    catch(Throwable whatever) {
      return null;
    }
  }

  /**
   * wraps a class in order to return a proper string representation
   * with respect to the flags
   */
  public ClassReflect(Class theClass) {
    this.theClass=theClass;
  }

  //the wrapped class
  Class theClass;

  /**
   * returns a formated string of the class content
   */
  public String toString() {
    return getString(theClass);
  }

  /**
   * methods relative to the wrapped class
   */
  public Class getWrappedClass() {
    return theClass;
  }
  public Constructor[] getConstructors() {
    return getConstructors(theClass);
  }
  public Field[] getFields() {
    return getFields(theClass);
  }
  public Method[] getMethods() {
    return getMethods(theClass);
  }
  public Class[] getHierarchy() {
    return getHierarchy(theClass);
  }
  public Class[] getInterfaces() {
    return theClass.getInterfaces();
  }

  /**
   * returns the class hierarchy of a given class
   */
  public static Class[] getHierarchy(Class current) {
    Vector result=new Vector();
    while(current!=null) {
      result.addElement(current);
      current=current.getSuperclass();
    }
    Class[] hierarchy=new Class[result.size()];
    result.copyInto(hierarchy);
    return hierarchy;
  }

  /**
   * says if the class data implements a given interface
   */
  public static boolean isImplementationOf(Class data,String classname) {
    if (data==null)
      return false;
    Class[] interfaces=data.getInterfaces();
    for(int i=0;i<interfaces.length;i++)
      if (interfaces[i].getName().equals(classname))
	return true;
    Class superclass=data.getSuperclass();
    if (superclass==null)
      return false;
    else
      return isImplementationOf(superclass,classname);
  }

  /**
   * says if the class data is an extension of the class parent
   */
  public static boolean isExtensionOf(Class data,Class parent) {
    return isExtensionOf(data,parent.getName());
  }

  /**
   * says if the class data is an extension of the className class name
   */
  public static boolean isExtensionOf(Class data,String className) {
    return (data!=null
	    && (data.getName().equals(className)
		|| isExtensionOf(data.getSuperclass(),className)));
  }

  /**
   * says if the class is the same as the className class name
   */
  public static boolean isSameClass(Class c,String className) {
    return c.getName().equals(className);
  }

  /**
   * manage the different modes
   */
  public static void setShorterNames(boolean b) {synchronized(STRING){ shorterNames=b; }}
  public static boolean isShorterNames() {synchronized(STRING){ return shorterNames; }}

  public static void setShorterTypes(boolean b) {synchronized(STRING){ shorterTypes=b; }}
  public static boolean isShorterTypes() {synchronized(STRING){ return shorterTypes; }}

  public static void enableSort(boolean b) {synchronized(STRING){ doSort=b; }}
  public static boolean isSortEnabled() {synchronized(STRING){ return doSort; }}

  public static void enableSortModifiers(boolean b) {synchronized(STRING){ sortModifiers=b; }}
  public static boolean isSortModifiers() {synchronized(STRING){ return sortModifiers; }}

  /**
   * manages the different filters
   */
  public static void setAccessFilter(int filter) {synchronized(STRING){ accessFlag=filter; }}
  public static int getAccessFilter() {synchronized(STRING){ return accessFlag; }}

  public static void setAbstractFilter(int filter) {synchronized(STRING){ abstractFlag=filter; }}
  public static int getAbstractFilter() {synchronized(STRING){ return abstractFlag; }}

  public static void setFinalFilter(int filter) {synchronized(STRING){ finalFlag=filter; }}
  public static int getFinalFilter() {synchronized(STRING){ return finalFlag; }}

  public static void setStaticFilter(int filter) {synchronized(STRING){ staticFlag=filter; }}
  public static int getStaticFilter() {synchronized(STRING){ return staticFlag; }}

  public static void setNativeFilter(int filter) {synchronized(STRING){ nativeFlag=filter; }}
  public static int getNativeFilter() {synchronized(STRING){ return nativeFlag; }}

  public static void setSynchronizedFilter(int filter) {synchronized(STRING){ synchronizedFlag=filter; }}
  public static int getSynchronizedFilter() {synchronized(STRING){ return synchronizedFlag; }}

  public static void setTransientFilter(int filter) {synchronized(STRING){ transientFlag=filter; }}
  public static int getTransientFilter() {synchronized(STRING){ return transientFlag; }}

  public static void setVolatileFilter(int filter) {synchronized(STRING){ volatileFlag=filter; }}
  public static int getVolatileFilter() {synchronized(STRING){ return volatileFlag; }}

  /**
   * returns the SortableClass of a class
   * according to the flags above
   */
  public static SortableClass getSortableClass(Class type) {
    SortableClass sc=new SortableClass(type,getString(type));
    if (sortModifiers)
      sc.setString(sc.getObjectString());
    return sc;
  }

  /**
   * returns the string representation of a class, in accord to
   * the above flags
   */
  public static String getString(Class type) {
    int dim=0;
    //special case: array
    while(type.isArray()) {
      dim++;
      type=type.getComponentType();
    }
    String className=(shorterTypes)?ResourceLoader.shortName(type):type.getName();
    //put the [] if it was an array
    if (dim>0) {
      StringBuffer buffer=new StringBuffer(className);
      for(int i=0;i<dim;i++)
	buffer.append("[]");
      return buffer.toString();
    }
    else
      return className;
  }

  /**
   * returns the modifiers of a class
   */
  public static StringBuffer getModifiers(Class theClass) {
    int mod=theClass.getModifiers();
    StringBuffer buffer=new StringBuffer(Modifier.toString(mod));
    if (buffer.length()>0)
      buffer.append(' ');
    if (!theClass.isInterface() && !theClass.isPrimitive())
      buffer.append("class ");
    return buffer;
  }

  /**
   * returns a sortable method with its strings in accord to the
   * above flags
   */
  public static SortableMethod getSortableMethod(Method method) {
    SortableMethod sm=new SortableMethod(method,getString(method));
    if (sortModifiers)
      sm.setString(sm.getObjectString());
    return sm;
  }

  /**
   * returns a string representation of the method according
   * to the above flags, but without modifiers
   */
  public static String getString(Method method) {
    StringBuffer buffer=new StringBuffer(128);

    //return type
    buffer.append(getString(method.getReturnType())).append(' ');
    //name
    if (!shorterNames)
      buffer.append(getString(method.getDeclaringClass())).append('.');

    //the parameters
    buffer.append(method.getName()).append('(');
    putString(method.getParameterTypes(),buffer);
    buffer.append(')');

    //the exception declaration
    Class[] excep=method.getExceptionTypes();
    if (excep.length>0) {
      buffer.append(" throws ");
      putString(excep,buffer);
    }

    return buffer.toString();
  }

  /**
   * returns the modifiers prefix for a method
   */
  public static StringBuffer getModifiers(Method method) {
    int mod=method.getModifiers();
    StringBuffer buffer=new StringBuffer(128);
    if (mod!=0)
      return buffer.append(Modifier.toString(mod)).append(' ');
    else
      return buffer;
  }

  /**
   * returns the SortableField of the field given in arguments,
   * according to the flags above
   */
  public static SortableField getSortableField(Field field) {
    SortableField sf=new SortableField(field,getString(field));
    if (sortModifiers)
      sf.setString(sf.getObjectString());
    return sf;
  }

  /**
   * returns the string representation of the field given
   * in argument, according to the flags above, but without modifiers
   */
  public static String getString(Field field) {
    int mod=field.getModifiers();
    StringBuffer buffer=new StringBuffer(64);
    Class type=field.getType();

    buffer.append(getString(type)).append(' ');
    if (!shorterNames)
      buffer.append(getString(field.getDeclaringClass())).append('.');
    buffer.append(field.getName());

    //gives the value when possible
    if (Modifier.isPublic(mod) && Modifier.isStatic(mod)) {
      try {
	Object obj=field.get(null);
	buffer.append('=');
	if (obj==null)
	  buffer.append("null");
	else if (type.equals(STRING) || type.equals(STRINGBUFFER))
	  buffer.append('\"').append(obj).append('\"');
	else if (type.equals(CHAR) || type.equals(CHARACTER))
	  buffer.append('\'').append(obj).append('\'');
	else if (type.equals(COLOR)) {
	  String colorname=getColorName((Color)obj);
	  if (colorname!=null)
	    buffer.append(getString(COLOR)).append('.').append(colorname);
	  else
	    buffer.append(obj);
	}
	else
	  buffer.append(obj);
      }
      catch(Exception e) {
	System.err.println(e);
      }
    }
    return buffer.toString();
  }

  /**
   * returns the string representation of a color, try the name representation
   */
  public static String getColorName(Color color) {
    if (color==null)
      return null;
    Object result=colorTable.get(color);
    if (result==null)
      return null;
    else
      return result.toString();
  }

  /**
   * returns the modifiers string of a field
   */
  public static StringBuffer getModifiers(Field field) {
    StringBuffer buffer=new StringBuffer(64);
    int mod=field.getModifiers();
    if (mod!=0)
      return buffer.append(Modifier.toString(mod)).append(' ');
    else
      return buffer;
  }

  /**
   * returns the SortableConstructor object of this constructor,
   * according to the flags above
   */
  public static SortableConstructor getSortableConstructor(Constructor constructor) {
    SortableConstructor sc=new SortableConstructor(constructor,getString(constructor));
    if (sortModifiers)
      sc.setString(sc.getObjectString());
    return sc;
  }

  /**
   * returns the string representation of the constructor given
   * in argument, according to the flags above, but without modifiers
   */
  public static String getString(Constructor constructor) {
    StringBuffer buffer=new StringBuffer(64);
    buffer.append(getString(constructor.getDeclaringClass()));

    //the parameters
    buffer.append('(');
    putString(constructor.getParameterTypes(),buffer);
    buffer.append(')');

    //the exceptions
    Class[] excep=constructor.getExceptionTypes();
    if (excep.length>0) {
      buffer.append(" throws ");
      putString(excep,buffer);
    }

    return buffer.toString();
  }

  /**
   * returns the modifiers of the given constructor
   */
  public static StringBuffer getModifiers(Constructor constructor) {
    StringBuffer buffer=new StringBuffer(64);
    int mod=constructor.getModifiers();
    if (mod!=0)
      return buffer.append(Modifier.toString(mod)).append(' ');
    else
      return buffer;
  }

  /**
   * fills the buffer with the string representations of the
   * list given in argument
   */
  public static void putString(Class[] list,StringBuffer buffer) {
    for(int i=0;i<list.length;i++) {
      buffer.append(getString(list[i]));
      if (i<list.length-1)
	buffer.append(',');
    }
  }

  /**
   * returns the short representation of the classes in argument
   */
  public static String[] getShortNames(Class[] list) {
    return getNames(list,true);
  }

  /**
   * returns short/long representation of the classes in argument
   */
  public static String[] getNames(Class[] list,boolean isShort) {
    //save flag
    boolean old=shorterTypes;
    String[] result=new String[list.length];
    shorterTypes=isShort;
    for(int i=0;i<result.length;i++)
      result[i]=getString(list[i]);
    //restore flag
    shorterTypes=old;
    return result;
  }

  /**
   * returns the short representation of the classes in argument
   */
  public static String[] getShortNames(String[] list) {
    String[] result=new String[list.length];
    for(int i=0;i<result.length;i++)
      result[i]=ResourceLoader.shortName(list[i]);
    return result;
  }

  /**
   * returns the long representation of the classes in argument
   */
  public static String[] getNames(Class[] list) {
    return getNames(list,false);
  }

  /**
   * convenient methods:+ modifiers
   */
  public static String[] getStrings(Method[] methods) {
    String[] strings=new String[methods.length];
    for(int i=0;i<strings.length;i++)
      strings[i]=getModifiers(methods[i]).append(getString(methods[i])).toString();
    return strings;
  }
  public static String[] getStrings(Field[] fields) {
    String[] strings=new String[fields.length];
    for(int i=0;i<fields.length;i++)
      strings[i]=getModifiers(fields[i]).append(getString(fields[i])).toString();
    return strings;
  }
  public static String[] getStrings(Constructor[] constructors) {
    String[] strings=new String[constructors.length];
    for(int i=0;i<constructors.length;i++)
      strings[i]=getModifiers(constructors[i]).append(getString(constructors[i])).toString();
    return strings;
  }
  public static String[] getStrings(Class[] classes) {
    String[] strings=new String[classes.length];
    for(int i=0;i<classes.length;i++)
      strings[i]=getModifiers(classes[i]).append(getString(classes[i])).toString();
    return strings;
  }
  public static SortableMethod[] getSortableMethods(Method[] methods) {
    SortableMethod[] strings=new SortableMethod[methods.length];
    for(int i=0;i<strings.length;i++)
      strings[i]=getSortableMethod(methods[i]);
    return strings;
  }
  public static SortableField[] getSortableFields(Field[] fields) {
    SortableField[] strings=new SortableField[fields.length];
    for(int i=0;i<fields.length;i++)
      strings[i]=getSortableField(fields[i]);
    return strings;
  }
  public static SortableConstructor[] getSortableConstructors(Constructor[] constructors) {
    SortableConstructor[] strings=new SortableConstructor[constructors.length];
    for(int i=0;i<constructors.length;i++)
      strings[i]=getSortableConstructor(constructors[i]);
    return strings;
  }
  public static SortableClass[] getSortableClasses(Class[] classes) {
    SortableClass[] strings=new SortableClass[classes.length];
    for(int i=0;i<classes.length;i++)
      strings[i]=getSortableClass(classes[i]);
    return strings;
  }

  /**
   * test: take or not with the filter and the modifier
   * filter of type yes/no/any
   */
  static boolean take(int filter,boolean modifier) {
    return (filter==ANY || (filter==YES && modifier) || (filter==NO && !modifier));
  }

  /**
   * test: take or not with the accessFlag and the modifier
   */
  static boolean take(int modifier) {
    return (accessFlag==ALLACCESS ||
	    (accessFlag==Modifier.PUBLIC && Modifier.isPublic(modifier)) ||
	    (accessFlag==Modifier.PROTECTED && Modifier.isProtected(modifier)) ||
	    (accessFlag==Modifier.PRIVATE && Modifier.isPrivate(modifier)) ||
	    (accessFlag==PACKAGE && (modifier & PACKAGE)==0));
  }

  /**
   * returns the corresponding fields with respect to the above filters
   */
  public static Field[] getFields(Class obj) {
    try {
      Field[] fields=obj.getDeclaredFields();
      Vector taken=new Vector(fields.length);
      for(int i=0;i<fields.length;i++) {
	int mod=fields[i].getModifiers();
	if (take(finalFlag,Modifier.isFinal(mod)) &&
	    take(staticFlag,Modifier.isStatic(mod)) &&
	    take(transientFlag,Modifier.isTransient(mod)) &&
	    take(volatileFlag,Modifier.isVolatile(mod)) &&
	    take(mod))
	  taken.addElement(fields[i]);
      }
      int n=taken.size();
      Field[] result=new Field[n];
      taken.copyInto(result);
      return result;
    }
    catch(SecurityException e) {
      return new Field[0];
    }
    catch(NullPointerException e) {
      System.err.println("getFields("+getString(obj)+"):"+e);
      return new Field[0];
    }
  }

  /**
   * returns the corresponding constructors with respect to the above filters
   */
  public static Constructor[] getConstructors(Class obj) {
    try {
      Constructor[] constructors=obj.getDeclaredConstructors();
      Vector taken=new Vector(constructors.length);
      for(int i=0;i<constructors.length;i++) {
	int mod=constructors[i].getModifiers();
	if (take(finalFlag,Modifier.isFinal(mod)) &&
	    take(mod))
	  taken.addElement(constructors[i]);
      }
      int n=taken.size();
      Constructor[] result=new Constructor[n];
      taken.copyInto(result);
      return result;
    }
    catch(SecurityException e) {
      return new Constructor[0];
    }
    catch(NullPointerException e) {
      System.err.println("getConstructors("+getString(obj)+"):"+e);
      return new Constructor[0];
    }
  }

  /**
   * returns the corresponding methods with respect to the above filters
   */
  public static Method[] getMethods(Class obj) {
    try {
      Method[] methods=obj.getDeclaredMethods();
      Vector taken=new Vector(methods.length);
      for(int i=0;i<methods.length;i++) {
	int mod=methods[i].getModifiers();
	if (take(abstractFlag,Modifier.isAbstract(mod)) &&
	    take(finalFlag,Modifier.isFinal(mod)) &&
	    take(nativeFlag,Modifier.isNative(mod)) &&
	    take(staticFlag,Modifier.isStatic(mod)) &&
	    take(synchronizedFlag,Modifier.isSynchronized(mod)) &&
	    take(mod))
	  taken.addElement(methods[i]);
      }
      int n=taken.size();
      Method[] result=new Method[n];
      taken.copyInto(result);
      return result;
    }
    catch(SecurityException e) {
      return new Method[0];
    }
    catch(NullPointerException e) {
      System.err.println("getMethods("+getString(obj)+"):"+e);
      return new Method[0];
    }
  }

  /**
   * tries to make a class from its name, possibly incomplete (without package prefix)
   * returns all found classes with the given name (several packages) if the flag is true
   */
  public static Class[] findClasses(String name,boolean all) {
    if (name.length()==0)
      return new Class[0];
    if (name.charAt(0)=='.')
      return findClasses(name.substring(1),all);
    if (name.charAt(name.length()-1)=='.')
      return new Class[0];

    Vector classes=(all)?new Vector():null;
    StringBuffer buffer=new StringBuffer(32);
    try {
      if (all)
	classes.addElement(Class.forName(name));
      else
	return new Class[] { Class.forName(name) };
    }
    catch(Throwable whatever){}

    //look for all the packages...
    String[] packages=PackageFinder.getPrefixedPackages();
    for(int i=0;i<packages.length;i++) {
      try {
	buffer.setLength(0);
	buffer.append(packages[i]).append(name);
	if (all) {
	  Class classFound=Class.forName(buffer.toString());
	  if (!classes.contains(classFound)) //because of "" packages and complete names...
	    classes.addElement(classFound);
	}
	else
	  return new Class[] { Class.forName(buffer.toString()) };
      }
      catch(Throwable whatever){}
    }
    if (classes==null)
      return new Class[0];
    Class[] result=new Class[classes.size()];
    classes.copyInto(result);
    return result;
  }

  /**
   * test
   */
  public static void main(String[] args) {
    try {
      Sorter.setListSeparator("\n  ");
      Sorter.setQuoted(false);
      int k=0;
      boolean all=true;
      if (args[0].equals("-one")) {
	k++;
	all=false;
      }
      System.out.println("Searching "+args[k]+" in "+PackageFinder.getPrefixedPackages().length+" packages...");
      Class[] classes=findClasses(args[k],all);
      setAccessFilter(ALLACCESS);
      System.out.println(Sorter.toString(getStrings(classes)));
      for(int i=0;i<classes.length;i++) {
	System.out.println("********** "+classes[i]);
	System.out.println("********** constructors *********\n"+
			   Sorter.toString(Sorter.sortI(getSortableConstructors(getConstructors(classes[i])))));
	System.out.println("********** fields *********\n"+
			   Sorter.toString(Sorter.sortI(getSortableFields(getFields(classes[i])))));
	System.out.println("********** methods *********\n"+
			   Sorter.toString(Sorter.sortI(getSortableMethods(getMethods(classes[i])))));
      }
    }
    catch(Exception x) {
      System.out.println(x);
      System.out.println("Usage: java adavid.reflect.ClassReflect [-one] className");
    }
  }
      

}
