/**
 * JavaFinder
 * This finds information on/in java sources
 * @version 1.2
 * @author Alexandre David
 */

/**
 * version history
 * 1.0: first version
 * 1.1: 990902
 *      removed findFile(..), put in getFile
 *      windoze bug fix and improve file reading
 * 1.2: 991011 added a filter for html
 * 1.3: 991107 filter removed
 */

package adavid.reflect;

import java.util.*;
import java.io.*;
import java.util.zip.*;
import adavid.util.*;

/**
 * JavaFinder
 * a java source finder: works on the class path and
 * the property java.source.path which contains
 * path to java source like CLASSPATH
 * The property is given to the program via the command line
 * with the option -Djava.source.path=...
 * Something strange is that if we change dynamically java.class.path,
 * the virtual machine does not take it into account!
 * Zip files may be given in argument, with a special clause [sub directory]
 * which says to search into the given sub directory ; this clause is ignored
 * for normal directories.
 */
public class JavaFinder {

  /**
   * the property definition
   */
  public final static String JAVA_PROPERTY_DEF="java.source.path";

  /**
   * the source directories, and the sub-directory definitions
   */
  static File[] sourceDir;
  static String[] subDir;

  /**
   * initializes the source directories
   */
  static {
    initJavaResource();
  }

  /**
   * initializes the java source resource
   * This may be useful if the user changes the value of
   * java.source.path and wants a dynamic update.
   */
  public static void initJavaResource() {
    String java_path=System.getProperty(JAVA_PROPERTY_DEF,".");
    StringTokenizer tokenizer=new StringTokenizer(java_path,File.pathSeparator);
    sourceDir=new File[tokenizer.countTokens()];
    subDir=new String[sourceDir.length];
    int i=0;
    while(tokenizer.hasMoreTokens()) {
      String dirName=tokenizer.nextToken();
      int index1=dirName.lastIndexOf("[");
      int index2=dirName.lastIndexOf("]");
      if (index1>0 && index2>index1) {
        subDir[i]=dirName.substring(index1+1,index2)+'/'/*File.separatorChar*/;
	dirName=dirName.substring(0,index1);
      }
      else
	subDir[i]="";
      sourceDir[i++]=new File(dirName);
    }
  }

  /**
   * returns the source of a class:java or jj
   */
  public static String getSource(Class aClass) {
    return getSource(aClass.getName());
  }
  public static String getjjSource(Class aClass) {
    return getjjSource(aClass.getName());
  }

  /**
   * returns the string content of the given filename ; null if error
   * @param filename : absolute class name
   * @param extension: extension of the file
   */
  public static String getFile(String filename,String extension) {
      //this is due to windoze, oh my god, so much
      //trouble for a such bad quality "OS", though I
      //hesitate naming this an "OS"
      String nameZip=filename.replace('.','/')+extension;
      //the test windoze is based on \ so I use directly '\\' instead of File.separatorChar
      String nameDir=(PackageFinder.windoze)?filename.replace('.','\\')+extension:nameZip;
      InputStream is=null;

      for(int i=0;i<sourceDir.length;i++)
	  if (sourceDir[i].isDirectory()) {
	      try {
		  File file=new File(sourceDir[i],nameDir);
		  if (file.exists()) {
		      //file => we have access to the file size
		      is=new FileInputStream(file);
		      byte[] data=new byte[(int)file.length()];
		      is.read(data,0,data.length);
		      is.close();
		      //now convert ascii -> char
		      char[] buffer=new char[data.length];
		      for(int j=0;j<data.length;j++)
			  buffer[j]=(char)data[j];
		      return new String(buffer);
		  }
	      }
	      catch(IOException ioe){}
	  }
	  else {
	      try {
		  ZipFile zip=new ZipFile(sourceDir[i]);
                  ZipEntry entry=zip.getEntry(subDir[i]+nameZip);
		  if (entry!=null) {
		      is=zip.getInputStream(entry);
		      break;
		  }
	      }
	      catch(Exception whatever){}
	  }
      //not found?
      if (is==null) {
	  is=ClassLoader.getSystemResourceAsStream(nameZip); 
	  //still not found?
	  if (is==null)
	      return null;
      }

      try{
	  byte[] data=new byte[16384];
	  int size=16384,used=0;
	  int nread;
	  //as we do not have access to the source size we
	  //read it by chunks
	  while((nread=is.read(data,used,size-used))>0) {
	      //System.out.println("nread="+nread+" size-used="+(size-used));
	      if (nread<size-used)
		  used+=nread;
	      else {
		  byte[] larger=new byte[size+=32768];//source generally small or very large
		  System.arraycopy(data,0,larger,0,used+=nread);
		  data=larger;
	      }
	  }
	  is.close();
	  char[] buffer=new char[used];
	  for(int j=0;j<used;j++)
	      buffer[j]=(char)data[j];
	  return new String(buffer);
      }
      catch(IOException ioe) {
	  System.err.println(ioe);
	  return null;
      }
  }

  /**
   * returns the source of a class; by absolute name ; null if not found, or error
   */
  public static String getSource(String className) {
    return getFile(className,".java");
  }

  /**
   * returns the jj source of a class; by absolute name ; null if not found, or error
   */
  public static String getjjSource(String className) {
    return getFile(className,".jj");
  }

  /**
   * returns the number of dimensions of the given name, i.e. occurences of '['
   */
  public static int countDimensions(String name) {
    int len=name.length();
    int dim=0,index=0;
    while(index<len && (index=name.indexOf('[',index)+1)>0)
      dim++;
    return dim;
  }

  /**
   * returns true if the declaration is identified
   * startp: points on '(', start: points on the function name
   */
  static boolean checkDeclaration(String source,int start,int startp,int len) {
    char c='\0';
    while(start>0 && (c=source.charAt(start))!='}'
	  && c!='{' && c!='=' && c!=',')
      start=goBack(source,start-1);
    if (c=='}') {
      // look for '{' just after
      startp++;
      while(startp<len && (c=source.charAt(startp))!='{' && c!=';')
	startp=jump(startp,source);
      return (c=='{');
    }
    else
      return (start==0);
  }

  /**
   * go back, jumps comments
   */
  public static int goBack(String source,int pos) {
    char c='\0';
    while(pos>=0 && (c=source.charAt(pos))==' ' || c=='\r' || c=='\t' || c=='\n')
      pos--;
    if (source.startsWith("public",pos-5) || source.startsWith("abstract",pos-7) || source.startsWith("native",pos-5))
      return 0;
    while(pos>=0 && (c=source.charAt(pos))!=' ' && c!='\r' && c!='\t' && c!='\n'
	  && c!=',' && c!='{' && c!='}' && c!='/')
      pos--;
    if (c=='\n') {
      int start=pos-1;
      while(start>0 && (c=source.charAt(start))!='\n' &&
	    (c!='/' || source.charAt(start-1)!='/'))
	start--;
      if (start>0 && c!='\n')
	start-=2;
      if (start<=0)
	start=0;
      if (start==0 || c!='\n')
	return start;
    }
    else if (c=='/' && pos>0 && source.charAt(pos-1)=='*') {
      pos-=2;
      while(pos>0 && (source.charAt(pos)!='*' || source.charAt(pos-1)!='/'))
	pos--;
      pos-=2;
      if (pos<0)
	pos=0;
    }
    return pos;
  }

  /**
   * returns the interval where the source of the constructor/method can
   * be found ; arguments are the name(short) and the arguments(long)
   * not supported: type / comment / []... name / comment / []...
   */
  public static Interval getSourcePos(String name,String[] args,String source) {
    //safe check
    if (source==null || name==null || args==null)
      return new Interval();

    //System.out.println(name+"("+Sorter.toString(args)+")");
    //take the dimension information
    int[] dim=new int[args.length];
    for(int i=0;i<args.length;i++) {
      dim[i]=countDimensions(args[i]);
      if (dim[i]>0)
	args[i]=args[i].substring(0,args[i].indexOf('['));
    } 
    //System.out.println(Sorter.toString(args));

    //make the short names equivalents
    String[] shortArgs=ClassReflect.getShortNames(args);
    //System.out.println(Sorter.toString(shortArgs));

    int start=0,end=0,len=source.length();
    boolean found=false;
    int nameLen=name.length();

    while(!found && start<len) {
      start=jump(start,source);
      if (start<0)
	start=len;
      //right name: constructor or method
      if (source.startsWith(name,start)
	  && start+nameLen<len
	  && !Character.isJavaLetterOrDigit(source.charAt(start+nameLen))) {
	//System.out.println(name);
	int start2=start;
	start+=nameLen-1;
	start=jump(start,source);
	//function identification
	if (start<len-1 && source.charAt(start)=='(' &&
	    (args.length!=0 || checkDeclaration(source,start2,start,len))) {
	  //System.out.println(name+"(");
	  boolean match=true;
	  //argument checking
	  for(int i=0;i<args.length && match;i++) {
	    int dimension=dim[i];
	    start=jump(start,source);
	    end=endOf(start+1,source);
	    String arg=(start<len && end<len)?source.substring(start,end):"";
	    //we have to corrent the needed [] after the name
	    while (end<len && source.charAt(end)=='[') {
	      dimension--;
	      end=endOf(jump(end,source),source);
	    }
	    //System.out.println(i+":"+arg+"=="+shortArgs[i]+" / "+source.substring(start-10,start+10));
	    match=(dimension>=0 && arg.equals((arg.indexOf('.')<0)?shortArgs[i]:args[i]));
	    //jump name if it is OK
	    if (match) {
	      start=jump(end,source);
	      if (dimension>0) {
		end=endOf(start,source);
		do {
		  end=endOf(end,source);
		  //System.out.println("test dim:"+source.substring(start,end+1));
		  if (end<len && source.charAt(end)=='[') {
		    dimension--;
		    end=jump(end,source);
		  }
		  else
		    match=false;
		} while(match && dimension>0);
		start=end;
	      }
	    }
	  }
	  //exact number of arg?
	  start=jump(start,source);
	  if (start<len && source.charAt(start)==')')
	    found=match;
	}
      }
      else
	start++;
    }
    
    if (found) {
      //we are on the right line
      char c='\0';
      end=start;
      start=source.lastIndexOf('\n',start);
      if (start<0)
	start=0;
      //try to find comments before this line
      int startbis=source.lastIndexOf('/',start);
      if (startbis>0) {
	int start2=-1;
	startbis--;
	c=source.charAt(startbis);
	if (c=='*') { //end of /* ... */
	  if (startbis>0)
	    start2=source.lastIndexOf("/*",startbis-1);
	  startbis++;//end of comments
	}
	else if (c=='/') {//begin of //...
	  start2=startbis;
	  startbis=source.indexOf('\n',startbis);//end of comments
	}
	if (start2>0) { //check that the comment belongs to the method
	  //between startbis(end of comments) and start(begin of line)
	  if (startbis<start) {
	    String between=source.substring(startbis,start);
	    if (between.indexOf(';')<0 && between.indexOf('}')<0)
	      start=start2;//validate if no declaration
	  }
	  else
	    start=start2;//validate
	}
      }
      
      //correct the start position to get the indentation
      start=source.lastIndexOf('\n',start);
      if (start<0)
	start=0;
      else
	start++;

      while (end<len && (c=source.charAt(end))!='{' && c!=';') end++;
      if (c=='{') {
	int bracket=1;
	end++;
	while(end<len && bracket>0) {
	  c=source.charAt(end);
	  if (c=='{')
	    bracket++;
	  else if (c=='}')
	    bracket--;
	  if (bracket>0)
	    end=jump(end,source);
	}
      }
      if (end<len)
	end++;
    }
    else
      return new Interval();

    return new Interval(start,end);
  }

  /**
   * jump until the next token, ignoring comments, blanks and ,
   * if we are on a token, then jump first to the end of this token
   */
  public static int jump(int pos,String string) {
    int len=string.length();
    if (pos>=len)
      return len;

    //we could have used endOf, but it is more efficient like this
    char c;
    int old=pos;
    while(pos<len && (c=string.charAt(pos))!=' '
	  && c!='\n' && c!='\t' && c!=',' && c!='(' && c!=')' && c!='[' && c!='\r'
	  && c!='{' && c!='}')
      pos++;
    if (old==pos)
      pos++;

    while(pos<len) {
      c=string.charAt(pos);
      if (c==' ' || c=='\n' || c=='\t' || c==',' || c=='\r')
	pos++;
      else if (c=='/') {
	pos++;
	if (pos<len) {
	  c=string.charAt(pos);
	  if (c=='*') { //jump until "*/"
	    pos=string.indexOf("*/",pos+1);
	    if (pos<0)
	      return len;
	    pos+=2;
	  }
	  else if (c=='/') { //jump one line
	    pos=string.indexOf('\n',pos+1);
	    if (pos<0)
	      return len;
	    pos++;
	  }
	  //else ignore
	}
	else
	  return pos;
      }
      else
	return pos;
    }
    return pos;
  }

  /**
   * finds the end of a token
   */
  public static int endOf(int pos,String string) {
    char c;
    int len=string.length();
    while(pos<len && (c=string.charAt(pos))!=' '
	  && c!='\n' && c!='\t' && c!=',' && c!='(' && c!=')' && c!='[' && c!='\r'
	  && c!='{' && c!='}')
      pos++;
    return pos;
  }

    /**
     * filter a normal source code to be printed correctly in HTML format
     * this is not optimized code, it is mainly intended to see if an
     * idea is worth
     *
    public static String filterToHTML(String source) {
	//System.out.println(source);
	int size=source.length();
	char[] txt=new char[size+2];
	StringBuffer buffer=new StringBuffer(2*size);
	int start=1,end=1,i=0;
	boolean comments=false;
	char c='\0';
	source.getChars(0,size,txt,1);
	txt[size+1]='\n';
	txt[0]='\n';
	size++;
	buffer.append("<html><body>");
	while(end<size) {
	    switch(txt[end++]) {
	    case '*':
		if (comments && txt[end]=='/') {
		    buffer.append(txt,start,end-start-1).append("</font></i>");
		    comments=false;
		    start=++end;
		}
		continue;
	    case ' ':
		if (txt[end]==' ' || txt[end]=='\t' || (end>1 && (txt[end-2]==' ' || txt[end-1]=='\t'))) {
		    buffer.append(txt,start,end-start-1).append("&nbsp;");
		    start=end;
		}
		continue;
	    case '\t':
		buffer.append(txt,start,end-start-1).append("&nbsp;&nbsp;&nbsp;&nbsp;");
		start=end;
		continue;
	    case '\n':
		buffer.append(txt,start,end-start-1).append("<br>");
		if (comments) {
		    while(end<size && ((c=txt[end])==' ' || c=='\t')) end++;
		    if (c=='*') {
			end++;
			if (txt[end]=='/') {
			    buffer.append("</font></i>");
			    comments=false;
			    end++;
			}
		    }
		}
		start=end;
		continue;
	    case '@':
		if (end+3<size && txt[end]=='s' && txt[end+1]=='e' && txt[end+2]=='e' &&
		    (txt[end+3]==' '||txt[end+3]=='\t')) {
		    end+=3;
		    ToADR:
		    while(end<size) {
			switch(txt[end++]) {
			case ' ':
			    buffer.append(txt,start,end-start-1).append("&nbsp;");
			    start=end;
			    continue ToADR;
			case '\t':
			    buffer.append(txt,start,end-start-1).append("&nbsp;&nbsp;&nbsp;&nbsp;");
			    start=end;
			    continue ToADR;
			default:
			    break ToADR;
			}
		    }
		    if (start+2<size && txt[start]=='<' &&
			(txt[start+1]=='a' || txt[start+1]=='A') &&
			(txt[start+2]==' ' || txt[start+2]=='\t')) {
			//anchor detected
			end=start+3;
			while(end+3<size && !(txt[end]=='<' && txt[end+1]=='/' &&
					      (txt[end+2]=='a' || txt[end+2]=='A') &&
					      txt[end+3]=='>')) end++;
			end+=4;//just jump
		    } else {
			//create anchor
			while(end<size && txt[end]!='\n') end++;
			buffer.append("<font color=\"#0000FF\"><a href=\"").append(txt,start,end-start).append("\">").append(txt,start,end-start).append("</a></font>");
			start=end;
		    }
		}
		continue;
	    case ';':
	    case '{':
	    case '}':
	    case '(':
	    case ')':
	    case '=':
	    case '[':
	    case ']':
	    case '!':
		buffer.append(txt,start,end-start-1).append("<b><font color=\"#0000FF\">").append(txt[end-1]).append("</font></b>");
		start=end;
		continue;
	    case '\"':
		buffer.append(txt,start,end-start-1).append("<b><font color=\"#F09024\">\"");
		start=end;
		while(end<size && (c=txt[end++])!='\"') {
		    if (c=='\\') end++;
		    if (txt[end-1]=='<') {
			buffer.append(txt,start,end-start-1).append("&lt;");
			start=end;
		    }
		}
		buffer.append(txt,start,end-start).append("</font></b>");
		start=end;
		continue;
	    case '\'':
		if (txt[end]=='<' && txt[end+1]=='\'') {
		    buffer.append(txt,start,end-start).append("&lt;'");
		    start=(end+=2);
		} else if (txt[end]=='\\' && (txt[end+1]=='\'' || txt[end+1]=='\"') && txt[end+2]=='\'') {
		    end+=3;
		    buffer.append(txt,start,end-start);
		    start=end;
		}
		continue;
	    case '/':
		switch(txt[end]) {
		case '/':
		    buffer.append(txt,start,end-start-1).append("<i><font color=\"#FF0070\">");
		    start=end-1;
		CommentJump:
		    while(end<size && txt[end++]!='\n') switch(txt[end-1]) {
		    case '<':
			buffer.append(txt,start,end-start-1).append("&lt;");
			start=end;
			continue CommentJump;
		    case '&':
			buffer.append(txt,start,end-start-1).append("&amp;");
			start=end;
			continue CommentJump;
		    }
		    buffer.append(txt,start,end-start-1).append("</font></i><br>");
		    start=end;
		    continue;
		case '*':
		    buffer.append(txt,start,end-start-1).append("<i><font color=\"#DD4070\">");
		    //while(end<size && txt[end]=='*') end++;
		    if (txt[end+1]=='*') end+=2; else end++;
		    start=end;
		    comments=true;
		}
		continue;
	    case '&':
		i=end;
		while(i<size && Character.isLetter(txt[i])) i++;
		if (txt[i]==';')
		    end=i+1;
		else {
		    buffer.append(txt,start,end-start-1).append("&amp;");
		    start=end;
		}
		continue;
	    case 's':
		if (end+11<size && !Character.isLetter(txt[end-2]) &&
		    txt[end]=='y' && txt[end+1]=='n' && txt[end+2]=='c' &&
		    txt[end+3]=='h' && txt[end+4]=='r' && txt[end+5]=='o' &&
		    txt[end+6]=='n' && txt[end+7]=='i' && txt[end+8]=='z' &&
		    txt[end+9]=='e' && txt[end+10]=='d' && !Character.isLetter(txt[end+11])) {
		    buffer.append(txt,start,end-start-1).append("<font color=\"#00BB00\">synchronized</font>");
		    start=(end+=11);
		} else if (end+5<size && !Character.isLetter(txt[end-2]) &&
			   txt[end]=='t' && txt[end+1]=='a' && txt[end+2]=='t' &&
			   txt[end+3]=='i' && txt[end+4]=='c' && !Character.isLetter(txt[end+5])) {
		    buffer.append(txt,start,end-start-1).append("<font color=\"#0088BB\">static</font>");
		    start=(end+=5);
		} else if (end+5<size && !Character.isLetter(txt[end-2]) &&
			   txt[end]=='w' && txt[end+1]=='i' && txt[end+2]=='t' &&
			   txt[end+3]=='c' && txt[end+4]=='h' && !Character.isLetter(txt[end+5])) {
		    buffer.append(txt,start,end-start-1).append("<font color=\"#00FF00\">switch</font>");
		    start=(end+=5);
		}
		continue;
	    case 'r':
		if (end+5<size && !Character.isLetter(txt[end-2]) &&
		    txt[end]=='e' && txt[end+1]=='t' && txt[end+2]=='u' &&
		    txt[end+3]=='r' && txt[end+4]=='n' && !Character.isLetter(txt[end+5])) {
		    buffer.append(txt,start,end-start-1).append("<font color=\"#00FF00\">return</font>");
		    start=(end+=5);
		}
		continue;
	    case 'n':
		if (end+5<size && !Character.isLetter(txt[end-2]) &&
		    txt[end]=='a' && txt[end+1]=='t' && txt[end+2]=='i' &&
		    txt[end+3]=='v' && txt[end+4]=='e' && !Character.isLetter(txt[end+5])) {
		    buffer.append(txt,start,end-start-1).append("<font color=\"#00FF77\">native</font>");
		    start=(end+=5);
		} else if (end+4<size && !Character.isLetter(txt[end-2]) &&
			   txt[end]=='e' && txt[end+1]=='w' && !Character.isLetter(txt[end+3])) {
		    buffer.append(txt,start,end-start-1).append("<font color=\"#00FF00\">new</font>");
		    start=(end+=3);
		}
		continue;
	    case 'e':
		if (end+6<size && !Character.isLetter(txt[end-2]) &&
		    txt[end]=='x' && txt[end+1]=='t' && txt[end+2]=='e' &&
		    txt[end+3]=='n' && txt[end+4]=='d' && txt[end+5]=='s'
		    && !Character.isLetter(txt[end+6])) {
		    buffer.append(txt,start,end-start-1).append("<font color=\"#0000AA\">extends</font>");
		    start=(end+=6);
		} else if (end+3<size && !Character.isLetter(txt[end-2]) &&
		    txt[end]=='l' && txt[end+1]=='s' && txt[end+2]=='e'
		    && !Character.isLetter(txt[end+3])) {
		    buffer.append(txt,start,end-start-1).append("<font color=\"#00FF00\">else</font>");
		    start=(end+=3);
		}
		continue;
	    case 't':
		if (end+4<size && !Character.isLetter(txt[end-2]) &&
		    txt[end]=='h' && txt[end+1]=='r' && txt[end+2]=='o' &&
		    txt[end+3]=='w') {
		    if (txt[end+4]=='s' && !Character.isLetter(txt[end+5])) {
			buffer.append(txt,start,end-start-1).append("<font color=\"#FF00FF\">throws</font>");
			start=(end+=5);
		    } else if (!Character.isLetter(txt[end+4])) {
			buffer.append(txt,start,end-start-1).append("<font color=\"#FF00FF\">throw</font>");
			start=(end+=4);
		    }
		} else if (end+3<size && !Character.isLetter(txt[end-2]) &&
			   txt[end]=='r' && txt[end+1]=='u' && txt[end+2]=='e'
			   && !Character.isLetter(txt[end+3])) {
		    buffer.append(txt,start,end-start-1).append("<b>true</b>");
		    start=(end+=3);
		} else if (end+2<size && !Character.isLetter(txt[end-2]) &&
			   txt[end]=='r' && txt[end+1]=='y' 
			   && !Character.isLetter(txt[end+2])) {
		    buffer.append(txt,start,end-start-1).append("<font color=\"#00FF00\">try</font>");
		    start=(end+=2);
		}
		continue;
	    case 'a':
		if (end+7<size && !Character.isLetter(txt[end-2]) &&
		    txt[end]=='b' && txt[end+1]=='s' && txt[end+2]=='t' &&
		    txt[end+3]=='r' && txt[end+4]=='a' && txt[end+5]=='c' &&
		    txt[end+6]=='t' && !Character.isLetter(txt[end+7])) {
		    buffer.append(txt,start,end-start-1).append("<font color=\"#FF6600\">abstract</font>");
		    start=(end+=7);
		}
		continue;
	    case 'f':
		if (end+6<size && !Character.isLetter(txt[end-2]) &&
		    txt[end]=='i' && txt[end+1]=='n' && txt[end+2]=='a' &&
		    txt[end+3]=='l' && txt[end+4]=='l' && 
		    txt[end+5]=='y' && !Character.isLetter(txt[end+6])) {
		    buffer.append(txt,start,end-start-1).append("<font color=\"#00FF00\">finally</font>");
		    start=(end+=6);
		} else if (end+4<size && !Character.isLetter(txt[end-2]) &&
		    txt[end]=='i' && txt[end+1]=='n' && txt[end+2]=='a' &&
		    txt[end+3]=='l' && !Character.isLetter(txt[end+4])) {
		    buffer.append(txt,start,end-start-1).append("<font color=\"#AA0066\">final</font>");
		    start=(end+=4);
		} else if (end+4<size && !Character.isLetter(txt[end-2]) &&
			   txt[end]=='a' && txt[end+1]=='l' && txt[end+2]=='s' &&
			   txt[end+3]=='e' && !Character.isLetter(txt[end+4])) {
		    buffer.append(txt,start,end-start-1).append("<b>false</b>");
		    start=(end+=4);
		}
		continue;
	    case 'c':
		if (end+4<size && !Character.isLetter(txt[end-2]) &&
		    txt[end]=='l' && txt[end+1]=='a' && txt[end+2]=='s' &&
		    txt[end+3]=='s' && !Character.isLetter(txt[end+4])) {
		    buffer.append(txt,start,end-start-1).append("<font color=\"#00FF77\">class</font>");
		    start=(end+=4);
		} else if (end+4<size && !Character.isLetter(txt[end-2]) &&
			   txt[end]=='a' && txt[end+1]=='s' && txt[end+2]=='e' &&
			   !Character.isLetter(txt[end+4])) {
		    buffer.append(txt,start,end-start-1).append("<font color=\"#00FF00\">case</font>");
		    start=(end+=4);
		} else if (end+7<size && !Character.isLetter(txt[end-2]) &&
			   txt[end]=='o' && txt[end+1]=='n' && txt[end+2]=='t' &&
			   txt[end+3]=='i' && txt[end+4]=='n' && 
			   txt[end+5]=='u' && txt[end+6]=='e' && !Character.isLetter(txt[end+7])) {
		    buffer.append(txt,start,end-start-1).append("<font color=\"#00FF00\">continue</font>");
		    start=(end+=7);
		} else if (end+4<size && !Character.isLetter(txt[end-2]) &&
			   txt[end]=='a' && txt[end+1]=='t' && txt[end+2]=='c' &&
			   txt[end+3]=='h' && !Character.isLetter(txt[end+5])) {
		    buffer.append(txt,start,end-start-1).append("<font color=\"#00FF00\">catch</font>");
		    start=(end+=4);
		}
		continue;
	    case 'i':
		if (end+1<size && !Character.isLetter(txt[end-2]) &&
		    txt[end]=='f' && !Character.isLetter(txt[end+1])) {
		    buffer.append(txt,start,end-start-1).append("<font color=\"#00FF00\">if</font>");
		    start=++end;
		} else if (end+9<size && !Character.isLetter(txt[end-2]) &&
		    txt[end]=='m' && txt[end+1]=='p' && txt[end+2]=='l' &&
		    txt[end+3]=='e' && txt[end+4]=='m' && txt[end+5]=='e' &&
		    txt[end+6]=='n' && txt[end+7]=='t' && txt[end+8]=='s' &&
		    !Character.isLetter(txt[end+9])) {
		    buffer.append(txt,start,end-start-1).append("<font color=\"#00FF00\">implements</font>");
		    start=(end+=9);
		} 
		continue;
	    case 'p':
		if (end+5<size && !Character.isLetter(txt[end-2]) &&
		    txt[end]=='u' && txt[end+1]=='b' && txt[end+2]=='l' && txt[end+3]=='i' && txt[end+4]=='c'
		    && !Character.isLetter(txt[end+5])) {
		    buffer.append(txt,start,end-start-1).append("<font color=\"#00FF33\">public</font>");
		    start=(end+=5);
		} else if (end+8<size && !Character.isLetter(txt[end-2]) &&
			   txt[end]=='r' && txt[end+1]=='o' && txt[end+2]=='t' && txt[end+3]=='e' && txt[end+4]=='c' &&
			   txt[end+5]=='t' && txt[end+6]=='e' && txt[end+7]=='d' && !Character.isLetter(txt[end+8])) {
		    buffer.append(txt,start,end-start-1).append("<font color=\"#AA5500\">protected</font>");
		    start=(end+=8);
		} else if (end+6<size && !Character.isLetter(txt[end-2]) &&
			   txt[end]=='r' && txt[end+1]=='i' && txt[end+2]=='v' && txt[end+3]=='a' && txt[end+4]=='t' &&
			   txt[end+5]=='e' && !Character.isLetter(txt[end+6])) {
		    buffer.append(txt,start,end-start-1).append("<font color=\"#FF0000\">private</font>");
		    start=(end+=6);
		}
		continue;
	    case '<':
		if (end+3<size &&
		    txt[end]=='p' &&
		    txt[end+1]=='r' &&
		    txt[end+2]=='e' &&
		    txt[end+3]=='>') {
		    end+=3;
		    PreLoop:
		    while(++end+5<size &&
			  !((c=txt[end])=='<'   &&
			    txt[end+1]=='/' &&
			    txt[end+2]=='p' &&
			    txt[end+3]=='r' &&
			    txt[end+4]=='e' &&
			    txt[end+5]=='>')) {
			switch(c) {
			case '\"':
			    buffer.append(txt,start,end-start).append("<b><font color=\"#F09024\">");
			    start=end++;
			    while(end<size && (c=txt[end++])!='\"')
				if (c=='\\') end++;
			    buffer.append(txt,start,end-start).append("</font></b>");
			    start=end--;
			    continue PreLoop;
			case '<':
			    if (txt[end+1]!='/') {
				i=end+1;
				while(i<size && Character.isLetter(txt[i])) i++;
				if (i>end+1 && txt[i]=='>')
				    end=i+1;
				else {
				    buffer.append(txt,start,end-start).append("&lt;");
				    start=end+1;
				}
			    }
			    continue PreLoop;
			case ';':
			case '(':
			case ')':
			case '{':
			case '}':
			case '=':
			case '[':
			case ']':
			case '!':
			    buffer.append(txt,start,end-start).append("<b><font color=\"#0000FF\">").append(c).append("</font></b>");
			    start=end+1;
			    continue PreLoop;
			case '\n':
			    if (comments) {
				i=end+1;
				while(i<size && ((c=txt[i])==' ' || c=='\t')) i++;
				if (c=='*') {
				    buffer.append(txt,start,i-start);
				    end=i;
				    start=i+1;
				}
			    }
			}
		    }
		}
		else if (comments && end+2<size &&
			 txt[end]=='h' &&
			 txt[end+1]=='r' &&
			 txt[end+2]=='>') {
		    buffer.append(txt,start,end-start-1).append("</font></i><hr><i><font color=\"#DD4070\"");
		    end+=3;
		    start=end;
		} else if (txt[end]!='/') {
		    i=end;
		    while(i<size && Character.isLetter(txt[i])) i++;
		    if (i>end && txt[i]=='>')
			end=i+1;
		    else {
			buffer.append(txt,start,end-start-1).append("&lt;");
			start=end;
		    }
		}
		continue;
	    }
	}
	buffer.append(txt,start,end-start).append("</body></html>");
	return buffer.toString();
  }
    */

  /**
   * test
   */
  public static void main(String[] args) {
    
    if (args.length==0)
      System.out.println("Usage: java adavid.reflect.JavaFinder className className ...");
    ClassReflect.setAccessFilter(ClassReflect.ALLACCESS);

    for(int i=0;i<args.length;i++) {
      String source=getSource(args[i]);
      if (source==null)
	source=getjjSource(args[i]);
      if (source==null)
	source=mocha.MochaGate.decompile(args[i]);
      if (source!=null) {
	try {
	  System.out.println("got source of "+args[i]);
	  java.lang.reflect.Constructor[] constructors=ClassReflect.getConstructors(Class.forName(args[i]));
	  for(int j=0;j<constructors.length;j++) {
	    String cs=SortableConstructor.getSource(source,constructors[j]);
	    if (cs!=null)
	      System.out.println(cs);
	    else
	      System.out.println("not found: "+constructors[j]);
	  }
	}
	catch(Exception e) {
	  System.out.println(e);
	}
      }
    }
  }

}
