/**
 * PackageFinder
 * This class is able to find all the packages of 
 * a java virtual machine. It searches through the
 * zip files and the directories. There is a little
 * bug when starting from the root and if there is
 * some link which goes backward. The process may not
 * terminate in this case. This class is used by
 * Assistant and constitutes its basis. It can
 * be executed independently as well.
 * @version 1.0.2
 * @author Alexandre David
 */

/**
 * version history
 * 1.0: first version
 * 1.0.1: 990818 identified bugs for windoze
 *        and forgotten / instead of File.separatorChar
 *        tried to fix bugs
 *        990829 changed backwards to '/' because
 *        it causes more problems under Windoze!
 * 1.0.2: 990902 even more windoze bug fixes!
 */

package adavid.reflect;

import java.util.*;
import java.io.*;
import java.util.zip.*;
import java.awt.*;
import adavid.util.*;

/**
 * PackageFinder
 * finds only the packages, does not record the classes,
 * java is too weak for this, the virtual machine crashes
 * if we make a Class for every class
 */
public class PackageFinder {

    /**
     * each package has its information represented as
     * a PackageResource class which can return information
     * on its classes.
     */
  private static PackageResource[] packageResources=null;

    /**
     * these labels can be used to display information while
     * processing the zip files and the directories. One has
     * to initialize them with the appropriate method.
     */
  static Label classPathLabel,pathLabel,packageLabel;

    /**
     * a local buffer for internal use
     */
  private static StringBuffer buffer=new StringBuffer(256);

    /**
     * a window to display information, can be set with
     * a method (see below)
     */
  static Window windowInfo;

    /**
     * the names of the packages, prefixPackages is basically
     * like packages with . at the end, which is used for completion.
     */
  private static String[] packages=null,prefixPackages=null;

    /**
     * setup windoze related problems correction
     * I do not use the test on os, but on the file separator
     * because the os could be Windoze 95/98/NT and I do not care
     */
    public final static boolean windoze=(File.separatorChar=='\\');

    /**
     * well, we need this hashtable only if windoze is here
     */
    private final static Hashtable dozeHash=(windoze)?new Hashtable():null;

    /**
     * prints information on the current searched path
     * if the label is set to a non null value
     */
  static void pathEcho(String path) {
    if (pathLabel!=null) {
      buffer.setLength(0);
      buffer.append("CLASSPATH: ").append(path);
      pathLabel.setText(buffer.toString());
      pathLabel.validate();//to avoid to paint twice
      pathLabel.paint(pathLabel.getGraphics());//to paint now
    }
  }

    /**
     * prints information on the current package if the label
     * is set to some non null value.
     */
  static void packEcho(PackageResource resource) {
    if (packageLabel!=null) {
      buffer.setLength(0);
      buffer.append("Package ").append(resource.pack);
      packageLabel.setText(buffer.toString());
      packageLabel.validate();//to avoid to paint twice
      packageLabel.paint(packageLabel.getGraphics());//to paint now
    }
  }

    /**
     * try to find a package which is located in path, in the zip entries entries
     * beginning with the path pathName in the package.
     * @param pathName sub package path
     * @param entries the zip entries of the zip class to look at
     * @param path the location of the zip file in fact
     */
  static PackageResource tryPackageResource(String pathName,ZipEntry[] entries,String path) {
    int sep=pathName.length()-1;
    for(int i=0;i<entries.length;i++) {
      String name=entries[i].getName();
      //System.out.println("from "+pathName+" ->"+name);
      if (!entries[i].isDirectory() &&
	  (sep<0 || (name.lastIndexOf('/'/*File.separatorChar*/)==sep &&
		     name.indexOf(pathName)==0)) &&
	  JMakefile.isExtFile(name,".class")) {
	try {
	  Class thisClass=Class.forName(JMakefile.shortName(
			   name.replace('/'/*File.separatorChar*/,'.')));
          //System.out.println("got "+pathName);
	  return new PackageResource(path,(sep<0)?"":pathName.substring(0,sep).replace('/'/*File.separatorChar*/,'.'));
	}
	catch(Throwable whatever) {}
      }
    }
    return null;
  }

  /*private static boolean notEmpty(Vector entryNames,String path,String pack,int n) {
    pathEcho(path);
    for(int i=0;i<n;i++) {
      String filename=(String)entryNames.elementAt(i);
      if (filename.lastIndexOf(File.separatorChar)==pack.length() && filename.indexOf(path)==0 
	  && JMakefile.isExtFile(filename,".class")) {
	try {
	  //sun.jws.Debugger.Agent makes crash
	  Class thisClass=Class.forName(JMakefile.shortName(
			    filename.replace(File.separatorChar,'.')));
	  return true;
	}
	catch(Throwable e) {
	  //System.err.println(e);
	}
      }
    }
    return false;
  }*/ //former alpha version

    /**
     * this function basically tests a path from a zip file entry list to
     * know if this path can be a valid package entry, i.e. if it contains valid classes.
     * @return true if the entry is recognized as a sub package entry
     * @param entries the zip entries to look at
     * @param path the sub package entry to validate
     * @param i where to begin the search, to speed up the process: the principle
     * is that we used this on a sorted list of entries so that assumption on the
     * location of a possible class is valid, which is we do not have to look at
     * the entire list of entries. Of course, we have to sort the list in n*log(n)
     * but we need to do this for each directory entry (or all possible entries
     * which can be sub-package entries).
     * @param pack the package name
     */
  private static boolean notEmpty(ZipEntry[] entries,String path,int i,String pack) {
    pathEcho(path);
    for(;i<entries.length;i++) {
      String filename=entries[i].getName();
      //System.out.println("path: "+path+", try "+filename+" for pack="+pack);
      if (filename.lastIndexOf('/'/*File.separatorChar*/)==pack.length()
       && filename.indexOf(path)==0) {
	if (JMakefile.isExtFile(filename,".class")) {
	  try {
	    //sun.jws.Debugger.Agent makes crash
            //System.out.println(filename+"?");
	    Class thisClass=Class.forName(JMakefile.shortName(
			      filename.replace('/'/*File.separatorChar*/,'.')));
            //System.out.println("OK:"+filename);
	    return true;
	  }
	  catch(Throwable e) {
	    //System.err.println(e);
	  }
	}
      }
      else
	return false;
    }
    return false;
  }

    /**
     * this is the equivalent of the previous notEmpty, but for directories
     * this time.
     * @return true if the path is recognized as a sub-package entry
     * @param path the sub package location, absolute path
     * @param dir the directory to look at
     * @param resources to register the packages which were found
     */
  private static boolean notEmpty(String path,File dir,Vector resources) {
    String[] list=dir.list();
    boolean empty=!dir.getPath().equals(".");
    //if the path is "." then it is not empty because this
    //depends on where it is executed, so we take this package
    //by default, so that it is always present

    String dirName=dir.getPath();
    int cutindex=path.length();
    if (dirName.length()>cutindex) {
	//the if is there because of different behaviours of java
	//or because a forgotten bug ;), oh yes in case of // in the path
	//but does not work for ///, who's got /// in his path???
	if (dirName.charAt(cutindex)==File.separatorChar)
	    dirName=dirName.substring(++cutindex);
	else
	    dirName=dirName.substring(cutindex+1); 
    }
    else
	dirName="";
    pathEcho(dirName);

    if (list==null) {
	System.err.println("Could not get list of "+dir);
	return false;
    }

    for(int i=0;i<list.length;i++) {
      String filename=list[i];
      File file=new File(dir,filename);
      //not very efficient here, I have to change this later
      String packPath=file.getPath().substring(cutindex);
      String pack=packPath.replace(File.separatorChar,'.');
      if (empty && !file.isDirectory()
	  && JMakefile.isExtFile(filename,".class")) {
	  pack=JMakefile.shortName(pack);
	  try {
	      //System.out.println("Trying "+pack+" path="+path+" packpath="+packPath); 
	      Class thisClass=Class.forName(pack);
	      empty=false;
	  }
	  catch(Throwable e) {
	    //System.err.println(e);
	    //if windoze the path is upper-cased!
	    if (windoze) {
		//try with lower cases!
		int lastdot=pack.lastIndexOf('.');
		if (lastdot>=0) {
                    String origpack=pack.substring(0,lastdot);
                    String newpack=origpack.toLowerCase();
                    pack=newpack+pack.substring(lastdot);
                    try {
                      Class thisClass=Class.forName(pack);
                      empty=false;
                      //ok now we record that we got
                      dozeHash.put(origpack,newpack);
                    }
		    catch(Throwable ee) { //no hope now!
		    }
		}
	    }
	}
      }
      else if (file.isDirectory()) {
	//URL url=new URL("systemresource:/FILE"+index+"/+/"+packPath);
	if (notEmpty(path,file,resources)) {
          PackageResource res;
          if (windoze) {
            String packname=(String)dozeHash.get(pack);
            //System.out.println("hash got:"+packname+" from "+pack);
            if (packname==null) packname=pack.toLowerCase();
            res=new PackageResource(path,packname);
          } else
            res=new PackageResource(path,pack);
	  resources.addElement(res);
	  packEcho(res);
	}
      }
    }
    return (!empty);
  }

    /**
     * this function is just to fix the Windoze problem that paths may
     * be of the form c:\...
     *
    public static String cleanUp(String path) {
	int index=path.indexOf(':');
	if (index<0) return path;
	else {
	    if (index+1<path.length())
		return path.substring(index+1);
	    else // path of the form c:
		return "";
	}
	}*/

    /**
     * return the packages of a system, tests before if
     * the packages were already computed and compute them
     * if needed
     */
  synchronized public static PackageResource[] getPackageResources() {
    if (packageResources==null)
      loadPackageResources();
    return packageResources;
  }

    /**
     * @return all the packages which matche a name
     * @param name the name of the package to match
     */
  synchronized public static PackageResource[] getPackageResources(String name) {
    if (packageResources==null)
      loadPackageResources();
    Vector resources=new Vector();
    for(int i=0;i<packageResources.length;i++)
      if(name.equals(packageResources[i].pack))
	resources.addElement(packageResources[i]);
    PackageResource[] result=new PackageResource[resources.size()];
    resources.copyInto(result);
    return result;
  }

    /**
     * @return the list of all the package names, similar to getPackageResources
     */
  synchronized public static String[] getPackages() {
    if (packages==null) {
      if (packageResources==null)
	loadPackageResources();
      packages=new String[packageResources.length];
      for(int i=0;i<packageResources.length;i++)
	packages[i]=packageResources[i].pack;
    }
    return packages;
  }

    /**
     * @return the list of all the prefixed package names (with .), similar to getPackages
     */
  synchronized public static String[] getPrefixedPackages() {
     if (prefixPackages==null) {
      if (packageResources==null)
	loadPackageResources();
      prefixPackages=new String[packageResources.length];
      for(int i=0;i<packageResources.length;i++)
	prefixPackages[i]=packageResources[i].getPrefixedPackage();
    }
    return prefixPackages;
  }
   
    /**
     * load the saved file containing the package resource description,
     * compares it to the executing JVM to see if an update is needed
     * and may update it. The update is performed if the file .PackageFinder.info
     * is not found.
     */
  private static void loadPackageResources() {
    try {
      String filename=System.getProperty("user.home")+'/'/*File.separatorChar*/+".PackageFinder.info";
      ObjectInputStream ins=new ObjectInputStream(new GZIPInputStream(new FileInputStream(filename)));
      String pathList=(String)ins.readObject();
      if (!pathList.equals(System.getProperty("java.class.path"))) {
	ins.close();
	throw new Exception(".PackageFinder.info has a different CLASSPATH");
      }
      packageResources=(PackageResource[])ins.readObject();
      ins.close();
    }
    catch(Exception whatever) {
      //System.out.println(whatever);
      updatePackageResources();
    }
  }

    /**
     * this function updates the packages and the file .PackageFinder.info
     */
  public static void updatePackageResources() {
    synchronized(buffer) {
      computePackageResources();
      //invalidate:
      packages=null;
      prefixPackages=null;
      try {
	String filename=System.getProperty("user.home")+File.separator+".PackageFinder.info";
	ObjectOutputStream ous=new ObjectOutputStream(new GZIPOutputStream(new FileOutputStream(filename)));
	ous.writeObject(System.getProperty("java.class.path"));
	ous.writeObject(packageResources);
	ous.flush();
	ous.close();
      }
      catch(Exception whatever) {
	System.err.println(whatever);
      }
    }
  }

    /**
     * this is the method to set the labels to display
     * information when reading information from the system.
     */
  public static void setLabels(Label classlabel,Label packlabel,Label pathlabel) {
    synchronized(buffer) {
      classPathLabel=classlabel;
      packageLabel=packlabel;
      pathLabel=pathlabel;
    }
  }	       

    /**
     * @return an array of unique strings got from the
     * tokenizer
     * @param tokenizer the tokenizer to get string from
     */
  public static String[] getUniqueStrings(StringTokenizer tokenizer) {
    Vector result=new Vector();
    while(tokenizer.hasMoreTokens()) {
      String string=tokenizer.nextToken();
      if (!result.contains(string))
	result.addElement(string);
    }
    int n=result.size();
    String[] strings=new String[n];
    result.copyInto(strings);
    //System.out.println(Sorter.toString(strings));
    return strings;
  }

    /**
     * this computes the packages found in
     * the virtual machine. This loads the zip file
     * and explores the directories. The previous functions
     * test if the entries are valid or not.
     */
  private static void computePackageResources() {
    String listpath=System.getProperty("java.class.path");
    //System.out.println(Sorter.toString(listpath));
    String[] classPaths=getUniqueStrings(new StringTokenizer(listpath,File.pathSeparator));
    Vector resources=new Vector();

    if (windoze) dozeHash.clear();

    if (windowInfo!=null)
      windowInfo.setVisible(true);

    for(int i=0;i<classPaths.length;i++) {
      String path=classPaths[i];
      File file=new File(path);

      if (file.exists()) {

	if (classPathLabel!=null) {
	  buffer.setLength(0);
	  buffer.append("class path ").append(i+1).append(File.separatorChar).append(classPaths.length).append(" : ").append(path);
	  classPathLabel.setText(buffer.toString());
	  classPathLabel.validate();//to avoid to paint twice
	  classPathLabel.paint(classPathLabel.getGraphics());//to paint now
	}

	if (file.isDirectory()) {
	  //URL url=new URL("systemresource:/FILE"+i+"/+/.");
	  if (notEmpty(path,file,resources)) {
	    PackageResource res=new PackageResource(path,"");
	    resources.addElement(res);
	    packEcho(res);
	  }
	}
	else {
	  try {
	    //ZipFile zip=new ZipFile(file);
	    /*ZipFile zip=PackageResource.getZipFile(file);
	    Vector entryNames=new Vector();
	    Vector entryDir=new Vector();
	    Enumeration entries=zip.entries();
	    while(entries.hasMoreElements()) {
	      ZipEntry entry=(ZipEntry)entries.nextElement();
	      if (entry.isDirectory())
		entryDir.addElement(entry.getName());
	      else
		entryNames.addElement(entry.getName());
	    }
	      
	    int nNames=entryNames.size();
	    int nDir=entryDir.size();
	    
	    for(int j=0;j<nDir;j++) {
	      String nameDir=(String)entryDir.elementAt(j);
	      //URL url=new URL("systemresource:/ZIP"+i+"/+/"+nameDir);
              String pack=nameDir.replace(File.separatorChar,'.');
	      pack=pack.substring(0,pack.length()-1);
	      if (notEmpty(entryNames,nameDir,pack,nNames)) {
		PackageResource res=new PackageResource(path,pack);
		resources.addElement(res);
		packEcho(res);
	      }
	    }*/ //this was the former version
	    //zip.close(); //no, since it is in a hash table

	      //zip files should not contain a drive name as c:\...
	    ZipEntry[] entries=PackageResource.getZipEntries(file);
	    Vector restPack=new Vector(),passed=new Vector();
	    if(entries!=null) {
	      for(int j=0;j<entries.length;j++)
		if (entries[j].isDirectory()) {
		  String nameDir=entries[j].getName();
                  //System.out.println("dir:"+nameDir);
		  String pack=nameDir.replace('/'/*File.separatorChar*/,'.');
		  passed.addElement(nameDir);
		  pack=pack.substring(0,pack.length()-1);
		  if (notEmpty(entries,nameDir,j+1,pack)) {
		    PackageResource res=new PackageResource(path,pack);
		    resources.addElement(res);
		    packEcho(res);
		  }
		}
		else {
		  String filename=entries[j].getName();
                  //System.out.println("file:"+filename);
		  int index=filename.lastIndexOf('/'/*File.separatorChar*/);
		  filename=filename.substring(0,index+1);
		  if (!restPack.contains(filename))
		    restPack.addElement(filename);
		}

	      int n=restPack.size();
	      for(int j=0;j<n;j++)
		if (!passed.contains(restPack.elementAt(j))) {
                  //System.out.println("not taken:"+restPack.elementAt(j)+" from "+file);
		  PackageResource res=tryPackageResource((String)restPack.elementAt(j),entries,path);
		  if (res!=null) {
		    //System.out.println("take "+restPack.elementAt(j));
		    resources.addElement(res);
		    packEcho(res);
		  }
		}
	    }

	  }
	  catch(Exception e) {
	    //System.err.println(e);
	  }
	}
      }
    }
    packageResources=new PackageResource[resources.size()];
    resources.copyInto(packageResources);
    
    if (windowInfo!=null)
      windowInfo.setVisible(false);
  }

    /**
     * this is the main function to test the
     * package independently
     */
  public static void main(String[] args) {
    Frame frame=new Frame("PackageFinder");
    frame.setLayout(new GridLayout(0,1));
    frame.add(classPathLabel=new Label());
    frame.add(pathLabel=new Label());
    frame.add(packageLabel=new Label());
    Dimension screen=frame.getToolkit().getScreenSize();
    int width=screen.width/3;
    int height=screen.height/7;
    frame.setBounds((screen.width-width)/2,(screen.height-height)/2,
		    width,height);
    windowInfo=frame;
    //System.out.println("Starting now...");

    if (args.length==0) {
      /*System.out.println("\nPackages:");
      String[] packs=getPrefixedPackages();
      for(int i=0;i<packs.length;i++)
	System.out.println(packs[i]);
      System.out.println();*/
      System.out.println("syntax: java adavid.util.PackageFinder "+
			 "[ complete_package_name [ complete/relative_class_name ] ]");

      int totalClasses=0;
      PackageResource[] rsrc=getPackageResources();
      for(int i=0;i<rsrc.length;i++) {
	int nbClass=rsrc[i].getClassNames().length;
	totalClasses+=nbClass;
	System.out.println(rsrc[i]+" ("+nbClass+" class"+((nbClass>1)?"es)":")"));
      }
      System.out.println(rsrc.length+" packages with "+totalClasses+" class"+((totalClasses>1)?"es.":"."));
    }
    else {
      PackageResource[] packs=getPackageResources(args[0]);
      if (packs.length==0)
	System.out.println("Package "+args[0]+" is not defined!");
      else {
	for(int j=0;j<packs.length;j++) {
	  PackageResource pack=packs[j];
	  System.out.println("\nPossible classes in "+pack);
	  String[] names=pack.getClassNames();
	  for(int i=0;i<names.length;i++)
	    System.out.println(names[i]);

	  Class[] top=pack.getTopClasses();
	  System.out.println("\nTop level class"+((top.length>1)?"es in ":" in ")+pack);
	  for(int i=0;i<top.length;i++)
	    System.out.println(top[i].getName());

	  if (args.length>1) {
	    System.out.println("\n"+args[1]+" has the following extensions in "+pack);
	    Class[] ext=pack.extensionsOf(args[1]);
	    for(int i=0;i<ext.length;i++)
	      System.out.println(ext[i].getName());
	  }
	}
      }
    }
    System.exit(0);
  }

}

