/**
 * LTreeCell
 * The cell of the node of a tree.
 * This is the representation class in fact,
 * since it extends JLabel.
 * @version 1.1
 * @author Alexandre David
 */

/**
 * version history:
 * 1.0: first version
 * 1.1 990829: use swing 1.1.1
 */

package adavid.swing;

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;

/**
 * LTreeCell
 * a cell of the LTreeNode
 */
public class LTreeCell extends JLabel implements MouseListener,FocusListener {
  
  /**
   * the default background color when the node
   * is selected
   */
  public static final Color ACTIVE=new Color(204,204,255);

  /**
   * the tree node which corresponds to this cell
   */
  LTreeNode node;

  /**
   * the popup menu of this cell, if any
   */
  JPopupMenu popup;

  /**
   * constructor: takes the node ; it is a hidden constructor because
   * the user is not expected to manipulate a cell directly
   */
  LTreeCell(LTreeNode node) {
    super(node.content.getText()+' ',
	  node.content.getIcon(false,true,false),
	  JLabel.LEFT);
    this.node=node;
    setForeground(Color.black);
    addMouseListener(this);
    addFocusListener(this);
    setToolTipText(node.content.getToolTip());
    setRequestFocusEnabled(true);
  }

  /**
   * the MouseListener implementation
   * call the LTListener ; collapse or expand
   */
  public void mouseClicked(MouseEvent e) {
    int iconW=(getIcon()==null)?0:getIcon().getIconWidth();
    if (e.getX()<iconW && e.getClickCount()==2) {
      node.saveViewport();
      if (node.isExpanded())
	node.collapse();
      else
	node.expand();
      node.restoreViewport();//needed because JScrollPane looses the view position
    }

    if (node.listeners!=null) {
      int n=node.listeners.size();
      for (int i=0;i<n;i++)
	if  (e.getClickCount()>1)
	  ((LTreeListener)node.listeners.elementAt(i)).nodeDoubleClicked(node,node.content);
	else
	  ((LTreeListener)node.listeners.elementAt(i)).nodeClicked(node,node.content);
    }
  }

  /**
   * implementation of MouseListener
   * select and update the icon ; open the popup if any
   */
  public void mousePressed(MouseEvent e) {
    node.unselect();
    node.select(this);
    node.updateIcon();
    grabFocus();

    if (e.isPopupTrigger() && popup!=null)
      popup.show(this,e.getX(),e.getY());
  }

  /**
   * returns true if the cell is selected
   */
  boolean isSelected() {
    return node.isSelected(this);
  }

  /**
   * useless methods of MouseListener
   */
  public void mouseReleased(MouseEvent e) {}
  public void mouseEntered(MouseEvent e) {}
  public void mouseExited(MouseEvent e) {}

  /**
   * the paint function ; overrides the JLabel paint function
   * in order to paint correctly the background!
   */
  public void paint(Graphics g) {
    Icon currentI = getIcon();
    Color bColor;
 
    /* set the background color */
    bColor=(isSelected())?ACTIVE:getParent().getBackground();
    g.setColor(bColor);
    Dimension dim=getPreferredSize();

    /* fill the background */
    if(currentI != null && getText() != null) {
      int offset = (currentI.getIconWidth() + getIconTextGap()/2);
      g.fillRect(offset, 0, dim.width-offset-1, dim.height-1);
      if (hasFocus()) {
        g.setColor(bColor.darker());
        g.drawRect(offset, 0, dim.width-offset-1, dim.height-1);
      }
    } else {
      g.fillRect(0, 0, dim.width-1, dim.height-1);
      if (hasFocus()) {
        g.setColor(bColor.darker());
        g.drawRect(0, 0, dim.width-1, dim.height-1);
      }
    }

    super.paint(g);
  }

  /**
   * the focus listener ; just repaint when the focus is gained or lost
   */
  public void focusGained(FocusEvent e) { repaint(); }
  public void focusLost(FocusEvent e)   { repaint(); }

}
