/**
 * LTreeNode
 * Label Tree Node
 * I create my own tree component because Java copies the nodes
 * without telling me when, and it copies the content of the nodes
 * as well, so when my other components refer to these nodes, well,
 * they are not valid because they have been replaced by new ones!
 * Imagine the time I've lost to understand where the problem came
 * from, and then to implement my own tree!!!
 * @version 1.1
 * @author Alexandre David
 */

/**
 * version history
 * 1.0: first version
 * 1.1 998029: use swing 1.1.1
 */

package adavid.swing;

import javax.swing.*;
import java.awt.*;
import java.util.Vector;
import java.awt.event.*;


public class LTreeNode extends JComponent implements MouseListener {

  /**
   * graphical information to draw the node (not the cell)
   */
  public final static int OFFSET=10;
  static final int[] pictx={6,3,0};
  static final int[] picty={0,6,0};

  /**
   * the parent of this node
   */
  LTreeNode parent;

  /**
   * the cell of this node and the selected cell ; the selected cell
   * is used by only the root node in fact
   */
  LTreeCell label,selection;

  /**
   * the children/leaves of this node
   */
  Vector children=new Vector();

  /**
   * the content of this node
   */
  LTreeContent content;

  /**
   * the dimension of this node ; has the sub nodes inside
   */
  Dimension nodeSize;

  /**
   * the listeners to be called in case of event
   */
  Vector listeners;

  /**
   * we save the view point of the JScrollPane if any because JScrollPane
   * looses this coordinate, so we restore it
   */
  Point saveView;

  /**
   * constructor: takes the content of the node ; the node
   * will be displayed with a cell
   */
  public LTreeNode(LTreeContent object) {
    this.parent=null;
    setLayout(null);
    content=object;
    if (object!=null) {
      label=new LTreeCell(this);
      nodeSize=new Dimension(label.getPreferredSize());
      super.add(label);
    }
    else nodeSize=new Dimension();
    setRequestFocusEnabled(true);
    addMouseListener(this);
  }

  /**
   * the node deletes itself if it is not the root
   */
  public void delete() {
    if (parent!=null)
      parent.removeNode(this);
  }

  /**
   * returns the parent node
   */
  public LTreeNode getParentNode() {
    return parent;
  }

  /**
   * returns true if this node is expanded
   */
  public boolean isExpanded() {
    if (label==null)
      return (getComponentCount()>0);
    else
      return (getComponentCount()>1);
  }

  /**
   * add a component or a node ; this is general and makes this
   * tree structure powerful and flexible
   */
  public Component add(Component c) {
    addSilent(c);
    expand();
    return c;
  }

  /**
   * same as above but without expand
   */
  public Component addSilent(Component c) {
    if (c instanceof LTreeNode) {
      LTreeNode node=(LTreeNode)c;
      node.parent=this;
      node.listeners=listeners;
    }
    super.add(c);
    children.addElement(c);
    return c;
  }


  /**
   * returns the preferred size of this node, which
   * has been computed with the sub nodes if the node is
   * expanded 
   */
  public Dimension getPreferredSize() {
    return new Dimension(nodeSize);
  }    

  /** 
   * remove a sub node attached to this node
   */
  public void removeNode(LTreeNode node) {
    remove(node);
    children.removeElement(node);
    hasChanged();
  }

  /**
   * removes all sub nodes attached to this node
   */
  public void removeAllNodes() {
    int n=children.size();
    for(int i=0;i<n;i++)
      remove((Component)children.elementAt(i));
    children.removeAllElements();
    hasChanged();
  }

  /**
   * expands this node and shows its sub nodes
   */
  public void expand() {
    removeAll();
    if (label!=null)
      super.add(label);
    int n=children.size();
    for(int i=0;i<n;i++)
      super.add((Component)children.elementAt(i));
    if (parent==null)
      hasChanged();
    else
      parent.expand();
  }

  /**
   * collapse this node and hide its sub nodes
   */
  public void collapse() {
    removeAll();
    if (label!=null)
      super.add(label);
    hasChanged();
  }

  /**
   * save the view port of the (maybe) parent viewport
   * this is useful because JScrollPane looses track of
   * the position of the view port
   */
  public void saveViewport() {
    if (parent==null) {
      if (getParent() instanceof JViewport)
	saveView=((JViewport)getParent()).getViewPosition();
    }
    else
      parent.saveViewport();
  }

  /**
   * restore the saved view port position
   */
  public void restoreViewport() {
    if (parent==null) {
      if (saveView!=null && getParent() instanceof JViewport)
	((JViewport)getParent()).setViewPosition(saveView);
    }
    else
      parent.restoreViewport();
  }

  /**
   * important function ; has to be called whenever a node is changed
   * and needs to be updated/displayed
   */
  public void hasChanged() {
    if (parent==null)
      setBounds();
    else
      parent.hasChanged();
    revalidate();
    invalidate();
    repaint();
  }

  /**
   * does the update job for an icon
   */
  void updateIcon() {
    if (label!=null)
      label.setIcon(content.getIcon((getComponentCount()>1),
				    (children.size()==0),
				    isSelected(label)));
  }

  /**
   * set and compute the bounds of the node ; take into account
   * the sub nodes if the node is expanded
   */
  void setBounds() {
    int h=0,w=0,first=0;
    if (label!=null) {
      first=1;
      label.setText(content.getText()+' ');
      label.setToolTipText(content.getToolTip());
      updateIcon();
      Dimension dim=label.getPreferredSize();
      label.setBounds(OFFSET,0,
		      w=dim.width+OFFSET,
		      h=Math.max(dim.height,OFFSET));
    }
    
    Component[] comp=getComponents();
    for(int i=first;i<comp.length;i++)
      if (comp[i] instanceof LTreeNode) {
	LTreeNode node=(LTreeNode)comp[i];
	node.setBounds();
	node.setLocation((label==null)?0:OFFSET,h);
	h+=node.getHeight();
	w=Math.max(w,node.getX()+node.getWidth());
      }
      else {
	Dimension dim=comp[i].getPreferredSize();
	int x=(label==null)?0:OFFSET;
	comp[i].setBounds(x,h,dim.width,dim.height);
	h+=dim.height;
	w=Math.max(w,dim.width+x);
      }

    setBounds(0,0,w,h);
    nodeSize.width=w;
    nodeSize.height=h;
  }

  /**
   * paint this component (node) ; basically paints a triangle if necessary (if
   * there is a content ; the cell is painted by the container function
   * paintComponents(..)
   */
  public void paintComponent(Graphics g) {
    //setBackground(getParent().getBackground());
    if (children.size()>0 && label!=null) { //has children && has place
      int shift=Math.max((label.getPreferredSize().height-6)/2,0);
      g.translate(0,shift);
      if (getComponentCount()>1) { //expanded
	g.setColor(Color.blue);
	g.fillPolygon(pictx,picty,3);
	g.setColor(Color.blue.brighter());
	g.drawPolyline(pictx,picty,3);
      }
      else { //collapsed
	g.setColor(Color.blue.darker());
	g.fillPolygon(picty,pictx,3);
	g.setColor(Color.blue);
	g.drawPolyline(picty,pictx,3);
      }
      g.translate(0,-shift);
    }
  }

  /**
   * set the popup of this node
   */
  public void setPopup(JPopupMenu popup) {
    if (label!=null) {
      label.add(popup);
      label.popup=popup;
      popup.setInvoker(label);
    }
  }

  /**
   * add a listener for this node
   */
  public void addLTreeListener(LTreeListener listener) {
    if (listeners==null)
      listeners=new Vector();
    listeners.addElement(listener);
  }

  /**
   * unselect his node
   */
  public void unselect() {
    if (parent==null) {
      LTreeCell cell=selection;
      selection=null;
      if (cell!=null)
	cell.node.updateIcon();
      repaint();
    }
    else
      parent.unselect();
  }
  
  /**
   * select a cell
   */
  void select(LTreeCell cell) {
    if (parent==null)
      selection=cell;
    else
      parent.select(cell);
  }

  /**
   * returns true if the cell is selected
   */
  boolean isSelected(LTreeCell cell) {
    if (parent==null)
      return (selection==cell);
    else
      return parent.isSelected(cell);
  }

  /**
   * collapse or expands if one clicks on the triangle
   */
  public void mousePressed(MouseEvent e) {
    if (label!=null
	&& e.getY()<=label.getPreferredSize().height
	&& e.getX()<=OFFSET) {
      saveViewport();
      if (isExpanded())
	collapse();
      else
	expand();
      restoreViewport();
    }
  }

  /**
   * useless methods of MouseListener
   */
  public void mouseClicked(MouseEvent e) {}
  public void mouseReleased(MouseEvent e) {}
  public void mouseEntered(MouseEvent e) {}
  public void mouseExited(MouseEvent e) {}

}
