/**
 * Bundle
 * basic (Resource)Bundle to initialize
 * classes; contains the basic functions
 * to build default or custom objects.
 * @version 1.2
 * @author Alexandre David
 */

/**
 * version history
 * 1.0: first version
 * 1.1 990801: added overloaded versions of init
 *      new feature introduced: init pure static
 *      classes + a final modifier added for the
 *      color table + static method loadImageFrom
 *      added so that an instance is not always
 *      necessary for loading an image.
 * 1.2 990819: revised code of loadImageFrom concerning
 *      reading bytes.
 *       990829: changed pathName to use '/' instead of
 *               File.separatorChar in order to solve paradoxically
 *               Windoze related problems
 *       990902: corrected loadImage once more
 */

package adavid.util;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import adavid.awt.*;
import java.lang.reflect.*;
import java.io.*;
import java.net.URL;

public class Bundle {

  /**
   * resource extensions to define how to initialise variables
   */
  public static final String ICON_DEF=".icon";
  public static final String NAME_DEF=".name";
  public static final String TEAR_DEF=".tearoff";
  public static String ICON_EXT=".gif";
  public static final String TRUE="true";
  public static final String FALSE="false";
  public static final String MNEMO_DEF=".mnemo";
  public static final String STATE_DEF=".state";
  public static final String COLOR_DEF=".color";
  public static final String PATH_NAME="path$";

  String iconPath=""; //overriden in declarations
  ResourceBundle bundle;
  Object data,actionHandler;
  DataResourceAccessible dataR;

  static final Hashtable colorTable=new Hashtable();
  public final static MediaTracker tracker=new MediaTracker(new Component(){});

  /**
   * initialization of the color table
   */
  static {
    try {
      Field[] field=Class.forName("java.awt.Color").getFields();
      for(int i=0;i<field.length;i++)
	if (field[i].getType().getName().equals("java.awt.Color")) {
	  try {colorTable.put(field[i].getName(),field[i].get(null));}
	  catch(Exception whatever){}
	}
    }
    catch(Exception whatever){}
  }
      

  /**
   * constructor with a class
   */
  public Bundle(Class target) {
    this(shortName(target.getName())+".properties",target);
  }

  /**
   * constructor: takes the name of the resource bundle, and
   * tries to locate the resource bundle
   */
  public Bundle(String bundleName,Class target) {
    String home=(target==null)?
	System.getProperty("user.home")+'/' /*File.separatorChar*/:
      pathName(target);
    //File.separatorChar avoided since it does not work properly
    //with getSystemResource

    try {
      URL url;
      url=ClassLoader.getSystemResource(bundleName);
      if (url==null)
	url=ClassLoader.getSystemResource(home+bundleName);
      if (url==null)
        bundle=new PropertyResourceBundle(new FileInputStream(bundleName));
      else
        bundle=new PropertyResourceBundle(url.openStream());
    }
    catch(Exception whatever) {
      bundle=null;
    }
  }

  /**
   * constructor: takes the resource bundle in argument
   */
  public Bundle(ResourceBundle bundle) {
    this.bundle=bundle;
  }

  /**
   * change the action handler
   */
  public void setActionHandler(Object actionHandler) {
    this.actionHandler=actionHandler;
  }

  /**
   * returns a resource definition
   *         or null is not found
   * characters '_' are converted to '.'
   */
  public String getResource(String def) {
    return getResource(def,null);
  }

  /**
   * returns a resource definition
   *         or default name is not found
   * characters '_' are converted to '.'
   */
  public String getResource(String def, String defaultResource) {
    try {
      //System.out.println("Try to access "+def.replace('_','.'));
      return bundle.getString(def.replace('_','.'));
    }
    catch(Exception whatever) {
      //System.out.println("Fail, return default="+defaultResource);
      return defaultResource;
    }
  }

  /**
   * convenient function: load a name definition
   */
  public String getNameResource(String def, String defaultValue) {
    return getResource(def+NAME_DEF,defaultValue);
  }

  /**
   * returns the color resource definition
   */
  public Color getColorResource(String def,Color defaultValue) {
    return getColor(getResource(def+COLOR_DEF),defaultValue);
  }

  /**
   * returns the bool resource definition
   */
  public boolean getBoolResource(String def,String defValue) {
    return getResource(def,defValue).toLowerCase().equals(TRUE);
  }

  /**
   * give the short name of a class
   */
  public static String shortName(Class c) {
    return shortName(c.getName());
  }

  /**
   * give the short name of a class name
   */
  public static String shortName(String longName) {
    return shortName(longName,'.');
  }

  /**
   * give the short name with separator as separator
   */
  public static String shortName(String longName,char separator) {
    int index=longName.lastIndexOf(separator);
    if (index<0)
      return longName;
    else if (index<longName.length()-1)
      return longName.substring(index+1);
    else
      return "";
  }

  /**
   * give the package-superclass name as path
   */
  public static String pathName(Class c) {
    return pathName(c.getName());
  }

  /**
   * give the package-superclass name as path
   */
  public static String pathName(String longName) {
    return pathName(longName,'.',true);
  }

  /**
   * give the path of longName, with separator as separator,
   * replace '.' by '/' instead of File.separatorChar if the flag is set, and the path is then of the
   * form "" or "name/name/"
   */
  public static String pathName(String longName,char separator,boolean replace) {
    int index=longName.lastIndexOf(separator);
    if (index<0)
      return "";
    else if (replace)
	return longName.substring(0,index+1).replace(separator,'/'/*File.separatorChar*/);
    else
      return longName.substring(0,index);
  }

  /**
   * returns the corresponding string resource with the name definition
   * for the string (the resource name) and its current value
   */
  public String getString(String name,String value) {
    return getNameResource(name,(value==null)?name:value);
  }

  /**
   * returns a color by its name or by its rgb definition in hexadecimal
   * or null in case of failure
   */
  public static Color getColor(String name,Color defaultColor) {
    if (name==null)
      return defaultColor;
    Object color=colorTable.get(name);
    if (color!=null)
      return (Color)color;
    //try hexa definition...
    int index=name.indexOf('#');
    if (index+1<name.length())
      name=name.substring(index);
    try {
      return new Color(Integer.parseInt(name,16));
    }
    catch(Exception whatever) {
      return defaultColor;
    }
  }

  /**
   * sets the location of a window so that it appears in the middle of the screen
   */
  public static void setLocation(Window win) {
    Dimension screenDim=win.getToolkit().getScreenSize();
    Dimension winDim=win.getSize();
    //This may seem stupid, but it does happen under Kaffe!
    if (screenDim.width==0)
	screenDim.width=1024;
    if (screenDim.height==0)
	screenDim.height=768;

    int x=(screenDim.width-winDim.width)/2;
    int y=(screenDim.height-winDim.height)/2;
    //setLocation does not work properly under Kaffe
    win.setLocation((x<0)?0:x,(y<0)?0:y);
  }

  /**
   * set location and make a window visible
   */
  public static void show(Window win) {
    setLocation(win);
    win.setVisible(true);
  }

  /**
   * position a window relative to its parent frame
   */
  public static void setLocation(Frame parent,Window win) {
    Rectangle r=parent.getBounds();
    Dimension d=win.getSize();
    Dimension s=win.getToolkit().getScreenSize();
    //This may seem stupid, but it does happen under Kaffe!
    if (s.width==0)
	s.width=1024;
    if (s.height==0)
	s.height=768;

    int x,y;

    //position in the middle of the parent if possible
    if (r.width>d.width)
      x=(r.width-d.width)/2+r.x;
    else
      x=r.x;
    if (r.height>d.height)
      y=(r.height-d.height)/2+r.y;
    else
      y=r.y;

    //check that win in inside the screen
    if (x+d.width>s.width)
      x=s.width-d.width;
    if (y+d.height>s.height)
      y=s.height-d.height;
    //setLocation does not work properly under Kaffe
    win.setLocation((x<0)?0:x,(y<0)?0:y);
  }

  /**
   * show a window relative to its parent frame
   */
  public static void show(Frame parent,Window win) {
    setLocation(parent,win);
    win.setVisible(true);
  }

  /**
   * return the image with the name definition def
   */
  public Image getImage(String def) {
    return loadImage(getResource(def+ICON_DEF,iconName(def)));
  }

  /**
   * change the icon path
   */
  public void setIconPath(String path) {
    if (path!=null)
      iconPath=path;
  }

  /**
   * returns the corresponding default name of a resource for an icon
   */
  public static String iconName(String def) {
    return def.replace('.','_')+ICON_EXT;
  }

  /**
   * easy solution:
   * return new Toolkit.getDefaultToolkit().getImage(ClassLoader.getSystemResource(iconPath+name));
   * but java crashes sometimes:
   Uncaught error fetching image:
   java.lang.NullPointerException
   at java.net.URLConnection.guessContentTypeFromStream(URLConnection.java)
   at sun.net.www.URLConnection.getContentType(URLConnection.java:126)
   at sun.awt.image.URLImageSource.getDecoder(URLImageSource.java:98)
   at sun.awt.image.InputStreamImageSource.doFetch(InputStreamImageSource.java:259)
   at sun.awt.image.ImageFetcher.fetchloop(ImageFetcher.java:151)
   at sun.awt.image.ImageFetcher.run(ImageFetcher.java:127)
   * so we do the job in another way: load the data and create an image from the data
   */
  //static version
    public static Image loadImageFrom(String completeName) throws IOException {
	if (completeName==null) return null;
	byte[] data=new byte[4096];
	int size=4096,used=0;
	InputStream is=ClassLoader.getSystemResource(completeName).openStream();
	int nread;
	while((nread=is.read(data,used,size-used))>0) {
	    if (nread<size-used) {
		used+=nread;
		//break; //do not break because is the input stream
		//is of type Inflater... we will get a bit less
	    } else {
		byte[] larger=new byte[size*=2];
		used+=nread;
		System.arraycopy(data,0,larger,0,used);
		data=larger;
	    }
	}
	is.close();
	Image image=Toolkit.getDefaultToolkit().createImage(data,0,used);
	if (waitImage(image))
	  return image;
	else
	  return null;
    }
  //intance version: add iconPath for the path
  public Image loadImage(String name) {
    try {
	return loadImageFrom(iconPath+name);
    }
    catch(Exception whatever) {
	//System.err.println(whatever+" on "+iconPath+name);
	return null;
    }
  }

  /**
   * wait until an image is ready to use
   */
  public static boolean waitImage(Image img) {
    if (img==null)
      return false;

    tracker.addImage(img,0);
    try {
      tracker.waitForID(0);
    }
    catch(InterruptedException ie) {
      return false;
    }
    finally {
      tracker.removeImage(img);
    }
    return true;
  }

  /**
   * returns the Button corresponding the given name,
   * use resource definitions to initialize the button
   */
  public Button getButton(String name) {
    String def=shortName(name,'_');
    Button button=new Button(getNameResource(name,def));
    button.setActionCommand(name);
    if (actionHandler!=null) {
      try {
	ActionManager.add(button,def,actionHandler);
	button.addActionListener(ActionManager.listener);
      }
      catch(Exception whatever){
      }
    }
    return button;
  }

  /**
   * same as getButton, for a check box
   */
  public Checkbox getCheckbox(String name) {
    String def=shortName(name,'_');
    Checkbox button=new Checkbox(getNameResource(name,def));
    button.setState(getBoolResource(name+STATE_DEF,FALSE));
    if (actionHandler!=null) {
      try {
	ActionManager.addItem(button,def,actionHandler);
	button.addItemListener(ActionManager.listener);
      }
      catch(Exception whatever){
      }
    }
    return button;
  }

  /**
   * configure a menu item with the resource definitions
   */
  public void setMenuItem(MenuItem item,String name,String def) {
    item.setActionCommand(name);
    String mnemoString=getResource(name+MNEMO_DEF,"");
    if (mnemoString.length()>0) {
      char mnemo=mnemoString.charAt(0);
      item.setShortcut(new MenuShortcut((int)mnemo));
    }
    if (!(item instanceof Menu || item instanceof CheckboxMenuItem) && actionHandler!=null) {
      try {
	ActionManager.add(item,def,actionHandler);
	item.addActionListener(ActionManager.listener);
      }
      catch(Exception whatever){
	//System.out.println(whatever);
      }
    }
  }

  /**
   * returns a configured menu item with the corresponding name definition
   */
  public MenuItem getMenuItem(String name) {
    String def=shortName(name,'_');
    MenuItem item=new MenuItem(getNameResource(name,def));
    setMenuItem(item,name,def);
    return item;
  }

  /**
   * same as getMenuItem, but for a check box menu item
   */
  public CheckboxMenuItem getCheckboxMenuItem(String name) {
    String def=shortName(name,'_');
    CheckboxMenuItem item=new CheckboxMenuItem(getNameResource(name,def));
    setMenuItem(item,name,def);
    item.setState(getBoolResource(name+STATE_DEF,FALSE));
    if (actionHandler!=null) {
      try {
	ActionManager.addItem(item,def,actionHandler);
	item.addItemListener(ActionManager.listener);
      }
      catch(Exception whatever){
	//System.out.println(whatever);
      }
    }
    return item;
  }

  /**
   * same as getMenuItem, but for a menu
   */
  public Menu getMenu(String name) {
    String def=shortName(name,'_');
    Menu menu=new Menu(getNameResource(name,def),
		       getBoolResource(name+TEAR_DEF,FALSE));
    setMenuItem(menu,name,def);
    return menu;
  }

  /**
   * returns a label with its name declared in the resource file
   */
  public Label getLabel(String name) {
    return new Label(getNameResource(name,shortName(name,'_')));
  }

  /**
   * return the corresponding popup menu
   */
  public PopupMenu getPopupMenu(String name) {
    return new PopupMenu(getNameResource(name,shortName(name,'_')));
  }

  /**
   * private access to the class to be initialized: get and set
   */
  Object get(Field var) throws IllegalAccessException {
    if (dataR!=null)
      return dataR.get(var);
    else
      return var.get(data);
  }
  void set(Field var,Object value) throws IllegalAccessException {
    if (dataR!=null)
      dataR.set(var,value);
    else
      var.set(data,value);
  }

  /**
   * the initialization function
   * called if data is not null => the class can be used only to load
   * resources in an easy way, with no initialization
   */
  public void init(Object data,Object actionHandler) {
    this.data=data;
    this.actionHandler=actionHandler;
    if (data!=null)
	init(data,data.getClass(),actionHandler);
  }

  /**
   * initialization for pure static classes, no instance required
   */
  public void init(Class dataClass) {
    if (dataClass!=null)
      init(null,dataClass,null);
  }

  /**
   * implementation of the initialization function
   */
  void init(Object data,Class dataClass,Object actionHandler) {
    if (dataClass!=null) {

      Field[] field=dataClass.getDeclaredFields();
      Vector menuList=new Vector();
      Vector itemList=new Vector();

      if (data instanceof DataResourceAccessible)
	dataR=(DataResourceAccessible)data;
      else
	dataR=null;

      for(int count=0;count<field.length;count++) {
	Field var=field[count];
	Class type=var.getType();

	//preliminary check before trying, otherwise it won't work, sure!
	int modifiers=var.getModifiers();
	if (!Modifier.isFinal(modifiers) &&
	    (dataR!=null || Modifier.isPublic(modifiers)) &&
	    (data!=null || Modifier.isStatic(modifiers))) {
	  try {
	    String name=var.getName();
	    String typeName=type.getName();
	    //System.out.println(typeName+" "+name);

	    if (typeName.equals("java.lang.String")) {
	      String value=(String)get(var);
	      String resrc=getNameResource(name,(value==null)?name:value);
	      if (name.indexOf(PATH_NAME)==0)
		iconPath=resrc;
	      set(var,resrc);
	    }
	    else if (typeName.equals("java.awt.Button")) {
	      Button button=getButton(name);
	      set(var,button);
	    }
	    else if (typeName.equals("java.awt.Menu")) {
	      Menu menu=getMenu(name);
	      set(var,menu);
	      tryToAttach(menuList,itemList,menu,name);
	      menuList.addElement(menu);
	      menuList.addElement(name);
	    }
	    else if (typeName.equals("java.awt.MenuItem")) {
	      MenuItem item=getMenuItem(name);
	      set(var,item);
	      tryToAttach(menuList,itemList,item,name);
	    }
	    else if (typeName.equals("java.awt.Color")) {
	      Color color=getColorResource(name,null);
	      if (color!=null)
		set(var,color);
	    }
	    else if (typeName.equals("java.awt.Image")) {
	      Image image=getImage(name);
	      if (image!=null)
		set(var,image);
	    }
	    else if (typeName.equals("java.awt.PopupMenu")) {
	      PopupMenu menu=getPopupMenu(name);
	      set(var,menu);
	      menuList.addElement(menu);
	      menuList.addElement(name);
	    }
	    else if (typeName.equals("java.awt.CheckboxMenuItem")) {
	      CheckboxMenuItem item=getCheckboxMenuItem(name);
	      set(var,item);
	      tryToAttach(menuList,itemList,item,name);
	    }
	    else if (typeName.equals("java.awt.Label")) {
	      set(var,getLabel(name));
	    }
	    else if (typeName.equals("java.awt.MenuBar")) {
	      MenuBar menuBar=new MenuBar();
	      set(var,menuBar);
	      menuList.addElement(menuBar);
	      menuList.addElement(name);
	    }
	  }
	  catch(Exception cannotAccess) {
	    //System.out.println(cannotAccess);
	    //cannotAccess.printStackTrace();
	  }
	}
      }
      
      // resolve the hierarchy and build the menu, if any
      if (menuList.size()>0) {
	int size=itemList.size()/2;
	for(int i=0;i<size;i++)
	  hasAttached(menuList,
		      (MenuItem)itemList.elementAt(2*i),
		      (String)itemList.elementAt(2*i+1));
      }
    }
  }

  /**
   * try to resolve the menu hierarchy, store if fails
   */
  void tryToAttach(Vector menuList,Vector storeList,MenuItem item,String name) {
    String parent=pathName(name,'_',false);

    if (parent.length()>0 && !hasAttached(menuList,item,parent)) {
      storeList.addElement(item);
      storeList.addElement(parent);
      //System.out.println("store "+name+" for parent "+parent);
    }
  }

  /**
   * try to resolve the menu hierarchy
   */
  boolean hasAttached(Vector menuList,MenuItem item,String parent) {
    int size=menuList.size()/2;

    for(int i=0;i<size;i++) {
      //System.out.println(parent+" =?= "+menuList.elementAt(2*i+1));
      if (parent.equals((String)menuList.elementAt(2*i+1))) {
	Object anchor=menuList.elementAt(2*i);
	if (anchor instanceof Menu)
	  ((Menu)anchor).add(item);
	else if (anchor instanceof MenuBar && item instanceof Menu)
	  ((MenuBar)anchor).add((Menu)item);
	else
	  System.err.println("Cannot attach "+item+" to "+anchor);
	return true;
      }
    }
    return false;
  }

}


