/**
 * CommandSorter
 * a class to handle the command line: options and so on
 * in an easy way
 * the command line is supposed to be written as:
 * java className -option optionarg -something argsomething onearg -hop hoparg anotherarg -hip -hipnorag ...
 * @version 1.0
 * @author Alexandre David
 */

/**
 * version history
 * 1.0:first version
 */

package adavid.util;

import java.util.Vector;

public class CommandSorter {

  /**
   * the command line
   */
  String[] args;

  /**
   * the option description
   */
  String option;

  /**
   * constructor: just take and store the
   * command line, and the option prefix definition
   */
  public CommandSorter(String[] args,String option) {
    this.args=args;
    this.option=option;
  }

  /**
   * default constructor: the prefixes are "-+"
   */
  public CommandSorter(String[] args) {
    this(args,"-+");
  }

  /**
   * returns true if the string "string" begins with
   * one of the characters in option, i.e. if it is an option
   * declaration
   */
  public boolean isOption(String string) {
    int n=option.length();
    if (string.length()<1 || n<1)
      return false;
    for(int i=0;i<n;i++)
      if (string.charAt(0)==option.charAt(i))
	return true;
    return false;
  }

  /**
   * returns the arguments which are not arguments of
   * an option
   */
  public String[] getArgs() {
    Vector list=new Vector();
    for(int i=0;i<args.length;i++) {
      while (i<args.length && isOption(args[i]))
	i++; //jump all options and arguments of options
      if (i<args.length) //something left
	list.addElement(args[i]);
    }
    String[] result=new String[list.size()];
    list.copyInto(result);
    return result;
  }

  /**
   * returns all arguments which are not options
   */
  public String[] getAllArgs() {
    Vector list=new Vector();
    for(int i=0;i<args.length;i++)
      if (!isOption(args[i]))
	list.addElement(args[i]);
    String[] result=new String[list.size()];
    list.copyInto(result);
    return result;
  }

  /**
   * returns all options found
   */
  public String[] getAllOptions() {
    Vector list=new Vector();
    for(int i=0;i<args.length;i++)
      if (isOption(args[i]))
	list.addElement(args[i]);
    String[] result=new String[list.size()];
    list.copyInto(result);
    return result;
  }

  /**
   * returns true if the argument is defined
   * options of the form -blabla=something are not counted
   */
  public boolean isDefined(String arg) {
    for(int i=0;i<args.length;i++)
      if (args[i].equals(arg))
	return true;
    return false;
  }

  /**
   * returns the argument of an option:
   * null if the option was not found, "" if no argument, "...something"
   * if the argument is found
   * the option should be complete, for example: "-o" for the option -o
   */
  public String getArg(String opt) {
    if (opt==null)
      return null;
    for(int i=0;i<args.length;i++)
      if (opt.equals(args[i])) {
	if (i==args.length-1 || isOption(args[i+1])) //no argument
	  return "";
	else
	  return args[i+1];
      }
    return null; //not found
  }

  /**
   * returns all arguments after a given option
   */
  public String[] getArgsAfter(String opt) {
    if (opt==null)
      return new String[0];
    for(int i=0;i<args.length;i++)
      if (opt.equals(args[i])) {
	i++;
	String[] result=new String[args.length-i];
	for(int j=0;i<args.length;j++,i++)
	  result[j]=args[i];
	return result;
      }
    return new String[0];//not found
  }

  /**
   * returns the definition of an option:
   * -option=something : getDefinition("-option") returns "something"
   * returns null if the option is not defined and "" if the definition
   * string is empty
   * -option = something ; -option= something ; -option =something are not recognized
   */
  public String getDefinition(String option) {
    if (option==null || option.length()==0)
      return null;
    for(int i=0;i<args.length;i++)
      if (args[i].indexOf(option)==0) {
	int defIndex=args[i].indexOf('=');
	if (defIndex<0)
	  defIndex=args[i].length();
	defIndex++;
	if (defIndex<args[i].length())
	  return args[i].substring(defIndex);
	else
	  return "";
      }
    return null;
  }

}
