/**
 * JDefaultResource
 * produces a file containing the resources
 * taken by default by JResource
 * in comment are the non defined resources which
 * do not have a default value
 * this approach is to avoid remembering the resources
 * of an application (or having them in a file which
 * should not be lost).  The application
 * has this information intrinsically and we extract it.
 * @version 1.0
 * @author Alexandre David
 */

/**
 * version history
 * 1.0: first distributed version
 */

package adavid.util;

import java.io.*;
import java.lang.reflect.*;
import adavid.reflect.ClassReflect;
import java.awt.Color;

public class JDefaultResource {

  /**
   * main program, needs the class name to find the default resource names.
   */
  public static void main(String[] args) {
    Class data=null;
    String filename;
    File file;

    //argument check
    if (args.length==0) {
      System.err.println("Usage:\n"+
			 "java adavid.util.JDefaultResource className [property_filename]");
      System.exit(1);
    }

    //class check
    try {
      data=Class.forName(args[0]);
    }
    catch(ClassNotFoundException e) {
      System.out.println(e);
      System.exit(2);
    }

    //optional property file/default file
    if (args.length>1)
      filename=args[1];
    else
      filename=JResource.shortName(args[0])+".properties";
    file=new File(filename);

    try {
      boolean append=false;
      if (file.exists()) {
	System.out.print(filename+" exists, append?(y/n)");
	char c=(char)System.in.read();
	append=(c=='y');
	if (append)
	  System.out.println("Warning : append properties to "+filename);
	else
	  System.out.println("Warning : overwrite "+filename);
      }
      PrintWriter out=new PrintWriter(new FileWriter(filename,append));
      Field[] field=data.getDeclaredFields();
      boolean canAccess=ClassReflect.isImplementationOf(data,"adavid.util.DataResourceAccessible");

      if (!append) {
	out.print("# File : "+filename+'\n'+
		  "# Default resource file produced by adavid.util.JDefaultResource\n"+
		  "# Change the values to have a customized resource file.\n"+
		  "# Uncomment and give definition to undefined resources.\n"+
		  "# Special case for button definition : if you set an empty name\n"+
		  "# and a valid icon, the button will have no name. An empty name with\n"+
		  "# an invalid icon will give the default name\n"+
		  "# The recognized resource definitions are:\n"+
		  "# * String name;                    //\"name\" is the default string\n"+
		  "# * String name=\"something\";        //\"something\" is the default string\n"+
		  "# * String path$comment=\"ho/\";      //ho/ becomes the new icon path by default, may be redefined\n"+
		  "#                                   //if it has no value, then the icon path does not change\n"+
		  "# * Icon name;                      //name.gif is the default icon\n"+
		  "# * JButton name;                   //\"name\" is the default name\n"+
		  "# * JRadioButton name;              //\"name\" is the default name\n"+
		  "# * JToggleButton name;             //\"name\" is the default name\n"+
		  "# * JMenu menu;                     //\"menu\" is the default name\n"+
		  "# * JMenu menu_submenu;             //\"submenu\" is the default name\n"+
		  "# * JMenuItem menu_submenu_item;    //\"item\" is the default name\n"+
		  "# * JCheckBoxMenuItem menu_checkbox;//\"checkbox\" is the default name\n"+
		  "# * JRadioButtonMenuItem menu_radio;//\"radio\" is the default name\n"+
		  "# * JSeparator menu_dummy;\n"+
		  "# * JRadioButtonMenuItem menu_zoup; //\"zoup\" is the default name\n"+
		  "# * RadioButtonMenuItem somewhere_button$somewhere2; //\"button\" is the default name of:\n"+
		  "#   - the JRadioButton attached somewhere\n"+
		  "#   - the RadioButtonMenuITem attached somewhere2\n"+
		  "# * ToggleButtonMenuItem somewhere_button$somewhere2; //same as previously with ToggleButton/MenuItem\n"+
		  "# * ToggleCheckButtonMenuItem ... with ToggleButton/CheckBoxMenuItem\n"+
		  "# * ButtonMenuItem ... with Button/MenuItem\n"+
		  "# * JLabel name;                    //\"name\" is the default name\n"+
		  "# * JMenuBar menuBar;\n"+
		  "# * JToolBar toolBar;\n"+
		  "# The following fill components do not have a resolution as powerful as before:\n"+
		  "# the parent component must have been defined before. One can complete the name by $comment\n"+
		  "# in order to have several of these components in the same container.\n"+
		  "# * Box.Filler where_glue;          //Box.createGlue()\n"+
		  "# * Box.Filler where_hglue;         //Box.createHorizontalGlue()\n"+
		  "# * Box.Filler where_vglue;         //Box.createVerticalGlue()\n"+
		  "# * Box.Filler where_h23;           //Box.createHorizontalStrut(23)\n"+
		  "# * Box.Filler where_v23;           //Box.createVerticalStrut(23)\n"+
		  "# * Box.Filler where_A34x21;        //Box.createRigidArea(new Dimension(34,21))\n"+
		  "# If the pressed icon resource is not found, then it is assigned to the rollover icon\n"+
		  "# The resources JButton, JToggleButton, JMenu, JMenuItem, JRadioButtonMenuItem, Box.Filler and JLabel\n"+
		  "# are hierarchical resources, and a hierarchy resolution is used to link them. One can use the JMenuBar and\n"+
		  "# the JToolBar names as root name.\n"+
		  "\n# AWT components may be initialized as well, if you use the\n"+
		  "# init function of Bundle, separately. You should not mix swing components\n"+
		  "# and awt components, they are not the same and for this reason the initializations\n"+
		  "# are not done in the same function. The possible object are:\n"+
		  "# Button, MenuBar, Menu, MenuItem, CheckboxMenuItem, String, PopupMenu, Label and Color\n"+
		  "# Color may be initialized by #rrggbb or a name of the Color class ; it works for JResource\n"+
		  "# they are initialized with the same principle, but buttons and labels are not attached.\n\n");

	out.print("# Flag to enable extended options.\n# To use the extended options:\n# * set ");
	out.println(JResource.EXTENDED_DEF+'='+JResource.TRUE);
	out.println("# * uncomment and define the commented options in this file");
	out.println(JResource.EXTENDED_DEF+'='+JResource.FALSE);
	out.println("# This flag is used only by ResourceLoader, not by Bundle\n");
      }

      for(int i=0;i<field.length;i++) {
	try {
	  Field var=field[i];
	  String type=var.getType().getName();
	  
	  if (!Modifier.isFinal(var.getModifiers()) &&
	      (canAccess || Modifier.isPublic(var.getModifiers()))) {
	    String name=var.getName();
	    
	    if (JResource.isString(type)) {
	      String value=null;
	      out.println("\n# String "+name);
	      if (name.indexOf(JResource.PATH_NAME)==0)
		out.println("# Special String definition: change the current icon path");
	      if (Modifier.isStatic(var.getModifiers()) &&
		   Modifier.isPublic(var.getModifiers()))
		value=(String)var.get(null);
	      else
		out.print("# "+name+
			    " is not public static : cannot access its instance value.\n#");
	      if (value==null)
		value=name;
	      out.println(name+JResource.NAME_DEF+'='+onOneLine(value));
	    }
	    else if (JResource.isIcon(type)) {
	      out.println("\n# Icon "+name);
	      out.println(name+JResource.ICON_DEF+'='+
			  name.replace('.','_')+JResource.ICON_EXT);
	    }
	    else if (JResource.isImageIcon(type)) {
	      out.println("\n# ImageIcon "+name);
	      out.println(name+JResource.ICON_DEF+'='+
			  name.replace('.','_')+JResource.ICON_EXT);
	    }
	    else if (type.equals("java.awt.Image")) {
	      out.println("\n# Image "+name);
	      out.println(name+JResource.ICON_DEF+'='+
			  name.replace('.','_')+JResource.ICON_EXT);
	    }
	    else if (JResource.isJPopupMenu(type)) {
	      out.println("\n# JPopupMenu "+name);
	      out.println(name+JResource.NAME_DEF+'='+name);
	      out.println(name+JResource.BORDER_PAINTED_DEF+'='+JResource.TRUE);
	    }
	    else if (JResource.isJLabel(type)) {
	      String shortName=JResource.shortName(name,'_');
	      String def=name.replace('_','.');
	      out.println("\n# JLabel "+name);
	      out.println(def+JResource.NAME_DEF+'='+shortName);
	      out.println(def+JResource.ICON_DEF+'='+
			  JResource.iconName(name));
	      out.println(def+JResource.TOOLTIP_DEF+'='+shortName);
	    }
	    else if (JResource.isJButton(type) || JResource.isJToggleButton(type) || JResource.isJRadioButton(type)) {
	      if (JResource.isJButton(type))
		out.println("\n# JButton "+name);
	      else if (JResource.isJToggleButton(type))
		out.println("\n# JToggleButton "+name);
	      else
		out.println("\n# JRadioButton "+name);

	      printButtonInfo(name,out,JResource.isJRadioButton(type),JResource.isJToggleButton(type),
			      JResource.isJButton(type)||JResource.isJToggleButton(type));
	    }
	    else if (JResource.isJMenu(type) || JResource.isJMenuItem(type)
		      || JResource.isJCheckBoxMenuItem(type)
		      || JResource.isJRadioButtonMenuItem(type)
		      || JResource.isToggleButtonMenuItem(type)
		      || JResource.isRadioButtonMenuItem(type)
		      || JResource.isToggleCheckButtonMenuItem(type)
		      || JResource.isButtonMenuItem(type)) {
	      out.println("\n# "+JResource.shortName(type)+" "+name);

	      String def2=null;
	      if (JResource.isToggleButtonMenuItem(type)
		  || JResource.isRadioButtonMenuItem(type)
		  || JResource.isToggleCheckButtonMenuItem(type)
		  || JResource.isButtonMenuItem(type)) {
		int index=name.indexOf('$');
		def2=name.substring(0,index);
		name=name.substring(index+1)+'_'+JResource.shortName(def2,'_');
	      }
	      String def=name.replace('_','.');

	      out.println(def+JResource.NAME_DEF+'='+JResource.shortName(def,'.'));
	      if (JResource.isJMenu(type))
		out.print('#');
	      out.println(def+JResource.ICON_DEF+'='+
			  JResource.iconName(name));
	      
	      if (JResource.isToggleButtonMenuItem(type))
		out.println(def+JResource.SELECTED_DEF+'='+
			    JResource.iconName(name+JResource.SELECTED_DEF));

	      if (JResource.isJCheckBoxMenuItem(type) || JResource.isToggleCheckButtonMenuItem(type))
		out.println(def+JResource.STATE_DEF+'='+JResource.FALSE);

	      out.println('#'+def+JResource.DISABLED_DEF+JResource.ICON_DEF+'='+
			  JResource.iconName(name+JResource.DISABLED_DEF));
	      if (JResource.isToggleButtonMenuItem(type))
		out.println('#'+def+JResource.DISABLED_SELECTED_DEF+'='+
			    JResource.iconName(name+JResource.DISABLED_SELECTED_DEF));

	      out.println('#'+def+JResource.ENABLED_DEF+'='+JResource.TRUE);
	      out.println('#'+def+JResource.MNEMO_DEF+'=');
	      if (def2!=null) {
		out.println("\n# associated button:");
		printButtonInfo(def2,out,JResource.isJRadioButtonMenuItem(type),JResource.isToggleButtonMenuItem(type),
				JResource.isToggleButtonMenuItem(type)||JResource.isButtonMenuItem(type)
				||  JResource.isToggleCheckButtonMenuItem(type));
	      }
	    }
	    else if (JResource.isJToolBar(type)) {
	      out.println("\n# JToolBar "+name);

	      out.println('#'+name+JResource.FLOATABLE_DEF+'='+JResource.TRUE);
	      out.println('#'+name+JResource.BORDER_PAINTED_DEF+'='+JResource.TRUE);
	    }
	    else if (type.equals("java.awt.MenuBar")) {
	      out.println("\n# MenuBar "+name+" is just initialized");
	    }
	    else if (type.equals("java.awt.Menu")) {
	      out.println("\n# Menu "+name);
	      out.println(name+Bundle.NAME_DEF+'='+Bundle.shortName(name,'_'));
	      out.println('#'+name+Bundle.MNEMO_DEF+'=');
	    }
	    else if (type.equals("java.awt.MenuItem")) {
	      out.println("\n# MenuItem "+name);
	      out.println(name+Bundle.NAME_DEF+'='+Bundle.shortName(name,'_'));
	      out.println('#'+name+Bundle.MNEMO_DEF+'=');
	    }
	    else if (type.equals("java.awt.CheckboxMenuItem")) {
	      out.println("\n# CheckboxMenuItem "+name);
	      out.println(name+Bundle.NAME_DEF+'='+Bundle.shortName(name,'_'));
	      out.println(name+Bundle.STATE_DEF+'='+Bundle.FALSE);
	      out.println('#'+name+Bundle.MNEMO_DEF+'=');
	    }
	    else if (type.equals("java.awt.Label")) {
	      out.println("\n# Label "+name);
	      out.println(name+Bundle.NAME_DEF+'='+Bundle.shortName(name,'_'));
	    }
	    else if (type.equals("java.awt.Button")) {
	      out.println("\n# Button "+name);
	      out.println(name+Bundle.NAME_DEF+'='+Bundle.shortName(name,'_'));
	    }
	    else if (type.equals("java.awt.Color")) {
	      out.println("\n# Color "+name);
	      String def="null";
	      try {
		Color color=(Color)var.get(null);
		if (color!=null) {
		  def=ClassReflect.getColorName(color);
		  if (def==null)
		    def='#'+Integer.toHexString(color.getRGB());
		}
	      }
	      catch(Exception x){}
	      out.println(name+Bundle.COLOR_DEF+'='+def);
	    }
	    else if (type.equals("java.awt.PopupMenu")) {
	      out.println("\n# PopupMenu "+name);
	      out.println(name+Bundle.NAME_DEF+'='+Bundle.shortName(name,'_'));
	    }
	  }
	}
	catch(Exception cannotAccess) {
	  System.out.println(cannotAccess);
	  //cannotAccess.printStackTrace();
	}
      }
      out.print("\n#End for "+data.getName()+"\n\n");
      out.flush();
      out.close();
    }
    catch(Exception whatever) { //should never happen, here for the PrintWriter constructor
      System.err.println(whatever);
      System.exit(5);
    }
  }

  static void printButtonInfo(String name,PrintWriter out,boolean isRadio,boolean isToggle,boolean hideOption) {
    String def=name.replace('_','.');
    String shortName=JResource.shortName(name,'_');
    out.println(def+JResource.NAME_DEF+'='+shortName);
    out.println(def+JResource.ICON_DEF+'='+
		JResource.iconName(name));
    if (!isToggle) out.print('#');
    out.println(def+JResource.SELECTED_DEF+JResource.ICON_DEF+'='+
		JResource.iconName(name+JResource.SELECTED_DEF));
    out.println(def+JResource.FOCUS_BORDER_DEF+'='+((isRadio)?JResource.FALSE:JResource.TRUE));
    out.println(def+JResource.TOOLTIP_DEF+'='+shortName);
    if (hideOption)
      out.println(def+JResource.HIDE_NAME_DEF+'='+JResource.TRUE);

    out.println('#'+def+JResource.PRESSED_DEF+JResource.ICON_DEF+'='+
		JResource.iconName(name+JResource.PRESSED_DEF));
    out.println('#'+def+JResource.RAISED_BORDER_DEF+'='+JResource.FALSE);
    out.println('#'+def+JResource.ROLLOVER_ENABLED_DEF+'='+JResource.FALSE);
    out.println('#'+def+JResource.ROLLOVER_DEF+JResource.ICON_DEF+'='+
		JResource.iconName(name+JResource.ROLLOVER_DEF));    
    out.println('#'+def+JResource.DISABLED_DEF+JResource.ICON_DEF+'='+
		JResource.iconName(name+JResource.DISABLED_DEF));
    out.println('#'+def+JResource.DISABLED_SELECTED_DEF+JResource.ICON_DEF+'='+
		JResource.iconName(name+JResource.DISABLED_SELECTED_DEF));
    out.println('#'+def+JResource.ROLLOVER_SELECTED_DEF+JResource.ICON_DEF+'='+
		JResource.iconName(name+JResource.ROLLOVER_SELECTED_DEF));
    out.println('#'+def+JResource.ENABLED_DEF+'='+JResource.TRUE);
    out.println('#'+def+JResource.FOCUS_PAINTED+'='+JResource.TRUE);
    out.println('#'+def+JResource.SELECTED_DEF+'='+JResource.FALSE);
    out.println('#'+def+JResource.BORDER_PAINTED_DEF+'='+JResource.FALSE);
    out.println('#'+def+JResource.MNEMO_DEF+'=');
  }

  public static String onOneLine(String text) {
    StringBuffer buffer=new StringBuffer(text.length());
    int i1=0,i2=0;
    while((i2=text.indexOf('\n',i1))>=0) {
      buffer.append(text.substring(i1,i2)).append("\\n");
      i1=i2+1;
    }
    buffer.append(text.substring(i1));
    return buffer.toString();
  }

}
