/**
 * JResource
 * a resource loader and initializer,
 * uses reflection to initialize the
 * class passed in argument
 * @version 1.1
 * @author Alexandre David
 */

/**
 * version history:
 * 1.0: first distributed version
 * 1.1 990829: use swing 1.1.1 now
 */

package adavid.util;

import javax.swing.*;
import java.awt.event.*;
import java.awt.*;
import java.net.URL;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;
import adavid.awt.*;
import adavid.swing.*;

/******************************************

syntax of the class passed in argument:

class Data {

 ### all members should be declared public, otherwise they may not be read or written

 static String name1;           => name1 is the resource name and the default value
 static String name2="default"; => here the default value is given
 static Icon name3;             => name3 is the resource name for an icon object
                            default is name3.gif

 ### String resources are defined as: resourceName.name=...
 ### Icon resources are defined as: resourceName.icon=...

 JMenu menu1;           => resource name + default name
                        => menu entry
 JMenuItem menu1_item1; => its menu item
 JMenu menu1_menu2;     => sub menu entry
 JMenuItem menu1_menu2_item => item of the sub menu

 JCheckBoxMenuItem jcbmi; => jcbmi is the default name + resource name ... no icon

 ### resource definitions:
 ### menu.name=... menu.mnemo=... for the menu and submenu
 ### menuitem.name=... menuitem.icon=... menuitem.mnemo=...
 ### the menu have the ActionListener given in argument to the constructor of
 ### JResource

 ### if no JMenu are defined , the file gives the structure:
 MenuBar= File Options
 File=Open Save Quit
 Open.action=menuItemOpen
 Open.icon=open.gif
 Save.action=menuItemSave
 Quit.action=menuItemQuit
 Options=Cool
 Cool=SubCool
 SubCool=jCheckBoxMenuItem

 ### definitions are static or not

 *******************************************/
public class JResource extends ResourceLoader {

  Class[] handlerArg;
  static JLabel titleLabel=null;
  static JProgressBar progressBar=null;

  /*******************
  ******* contructors
  ********************/
  public JResource(Object data) {
    this(data,null);
  }

  public JResource(Object data, ActionListener listener) {
    this(data,listener,(data!=null)?shortName(data.getClass())+".properties":null);
  }

  public JResource(Object data, ActionListener listener, String bundleName) {
    this(data,listener,bundleName,(listener==null)?data:listener);
  }

  public JResource(Object data, ActionListener listener, String bundleName, Object actionHandler) {
    super(bundleName,(data==null)?null:data.getClass());
    initialize(data,listener,actionHandler);
  }

  public JResource(Object data, ActionListener listener, ResourceBundle bundle) {
    this(data,listener,bundle,(listener==null)?data:listener);
  }

  public JResource(Object data, ActionListener listener, ResourceBundle bundle, Object actionHandler) {
    super(bundle);
    initialize(data,listener,actionHandler);
  }

  /**
   * the initialization function
   * called if data is not null => the class can be used only to load
   * resources in an easy way, with no initialization
   */
  public void initialize(Object data,ActionListener listener) {
    initialize(data,listener,(listener==null)?data:listener);
  }
  public void initialize(Object data,ActionListener listener,Object actionHandler) {
    this.listener=listener;
    this.data=data;
    this.actionHandler=actionHandler;
    if (data!=null) {

      Field[] field=data.getClass().getDeclaredFields();
      Vector menuList=new Vector();
      Vector itemList=new Vector();
      ButtonGroup group=null,groupmenu=null;
      AbstractButton lastButton=null,lastItem=null;
      int buttonCount=0;

      if (titleLabel!=null)
	titleLabel.setText(data.getClass().getName());
      if (progressBar!=null) {
	progressBar.setMinimum(0);
	progressBar.setValue(0);
	progressBar.setMaximum(field.length);
      }

      if (data instanceof DataResourceAccessible)
	dataR=(DataResourceAccessible)data;
      else
	dataR=null;

      for(int count=0;count<field.length;count++) {
	if (progressBar!=null) {
	  progressBar.setValue(count+1);
	  try{progressBar.paintImmediately(progressBar.getBounds());}
	  catch(Exception e){/* an exception is raised the first time, god knows why... */}
	}

	Field var=field[count];
	Class type=var.getType();

	//preliminary check before trying, otherwise it won't work, sure!
	int modifiers=var.getModifiers();
	if (!Modifier.isFinal(modifiers) && (dataR!=null || Modifier.isPublic(modifiers))) {
	  try {
	    String name=var.getName();
	    String typeName=type.getName();
	    //System.out.println(typeName+" "+name);
	    if (isString(typeName)) {
	      String value=(String)get(var);
	      String resrc=getNameResource(name,(value==null)?name:value);
	      if (name.indexOf(PATH_NAME)==0)
		iconPath=resrc;
	      set(var,resrc);

	      buttonCount=0;
	    }
	    else if (isIcon(typeName) || isImageIcon(typeName)) {
	      ImageIcon icon=getIcon(name);
	      if (icon!=null)
		set(var,icon);

	      buttonCount=0;
	    }
	    else if (isJButton(typeName)) {
	      JButton button=getJButton(name);
	      set(var,button);
	      tryToResolve(menuList,itemList,button,name);
	      buttonCount=0;
	    }
	    else if (isJToggleButton(typeName)) {
	      JToggleButton button=getJToggleButton(name);
	      set(var,button);

	      tryToResolve(menuList,itemList,button,name);
	      buttonCount++;
	      if (buttonCount==2) {
		group=new ButtonGroup();
		group.add(lastButton);
	      }
	      if (buttonCount>1)
		group.add(button);
	      lastButton=button;
	    }
	    else if (isJRadioButton(typeName)) {
	      JRadioButton button=getJRadioButton(name);
	      set(var,button);

	      tryToResolve(menuList,itemList,button,name);
	      buttonCount++;
	      if (buttonCount==2) {
		group=new ButtonGroup();
		group.add(lastButton);
	      }
	      if (buttonCount>1)
		group.add(button);
	      lastButton=button;
	    }
	    else if (isJMenu(typeName)) {
	      JMenu menu=getJMenu(name);
	      set(var,menu);

	      tryToResolve(menuList,itemList,menu,name);
	      menuList.addElement(menu);
	      menuList.addElement(name);
	      buttonCount=0;
	    }
	    else if (isJMenuItem(typeName)) {
	      JMenuItem item=getJMenuItem(name);
	      set(var,item);

	      tryToResolve(menuList,itemList,item,name);
	      buttonCount=0;
	    }
	    else if (isColor(typeName)) {
	      Color color=getColorResource(name,null);
	      if (color!=null)
		set(var,color);
	      buttonCount=0;
	    }
	    else if (isJPopupMenu(typeName)) {
	      JPopupMenu menu=getJPopupMenu(name);
	      set(var,menu);

	      //tryToResolve(menuList,itemList,menu,name); //do not attach a popup menu
	      menuList.addElement(menu);
	      menuList.addElement(name);
	      buttonCount=0;
	    }
	    else if (isToggleButtonMenuItem(typeName)) {
	      int index=name.lastIndexOf('$');
	      String buttonName=name.substring(0,index);
	      String itemName=name.substring(index+1)+'_'+shortName(buttonName,'_');

	      JToggleButton button=getJToggleButton(buttonName);
	      tryToResolve(menuList,itemList,button,buttonName);

	      /*if (button.getIcon()!=null) {
		button.setText("");
		button.setMargin(new Insets(0,0,0,0));
		}*/ //this is now done on all JButton and JToggleButton

	      ToggleButtonMenuItem item=getToggleButtonMenuItem(itemName,button);
	      set(var,item);
	      tryToResolve(menuList,itemList,item,itemName);

	      buttonCount++;
	      if (buttonCount==2) {
		group=new ButtonGroup();
		group.add(lastButton);
		if (lastItem!=null) {
		  groupmenu=new ButtonGroup();
		  groupmenu.add(lastItem);
		}
	      }
	      if (buttonCount>1) {
		group.add(button);
		if (groupmenu!=null)
		  groupmenu.add(item);
	      }
	      lastItem=item;
	      lastButton=button;
	    }
	    else if (isButtonMenuItem(typeName)) {
	      int index=name.lastIndexOf('$');
	      String buttonName=name.substring(0,index);
	      String itemName=name.substring(index+1)+'_'+shortName(buttonName,'_');

	      JButton button=getJButton(buttonName);
	      tryToResolve(menuList,itemList,button,buttonName);

	      /*if (button.getIcon()!=null) {
		button.setText("");
		button.setMargin(new Insets(0,0,0,0));
		}*/ // this is now done on all JButton and JToggleButton

	      ButtonMenuItem item=getButtonMenuItem(itemName,button);
	      set(var,item);
	      tryToResolve(menuList,itemList,item,itemName);

	      buttonCount=0;
	    }
	    else if (isRadioButtonMenuItem(typeName)) {
	      int index=name.lastIndexOf('$');
	      String buttonName=name.substring(0,index);
	      String itemName=name.substring(index+1)+'_'+shortName(buttonName,'_');

	      JRadioButton button=getJRadioButton(buttonName);
	      tryToResolve(menuList,itemList,button,buttonName);

	      /*if (button.getIcon()!=null) {
		button.setText("");
		button.setMargin(new Insets(0,0,0,0));
		}*/ //this is not done for radio buttons

	      RadioButtonMenuItem item=getRadioButtonMenuItem(itemName,button);
	      set(var,item);
	      tryToResolve(menuList,itemList,item,itemName);

	      buttonCount++;
	      if (buttonCount==2) {
		group=new ButtonGroup();
		group.add(lastButton);
		if (lastItem!=null) {
		  groupmenu=new ButtonGroup();
		  groupmenu.add(lastItem);
		}
	      }
	      if (buttonCount>1) {
		group.add(button);
		if (groupmenu!=null)
		  groupmenu.add(item);
	      }
	      lastItem=item;
	      lastButton=button;
	    }
	    else if (isJRadioButtonMenuItem(typeName)) {
	      JRadioButtonMenuItem item=getJRadioButtonMenuItem(name);
	      set(var,item);

	      tryToResolve(menuList,itemList,item,name);
	      buttonCount++;
	      if (buttonCount==2) {
		group=new ButtonGroup();
		group.add(lastButton);
	      }
	      if (buttonCount>1)
		group.add(item);
	      lastButton=item;
	    }
	    else if (isJCheckBoxMenuItem(typeName)) {
	      JCheckBoxMenuItem item=getJCheckBoxMenuItem(name);
	      set(var,item);

	      tryToResolve(menuList,itemList,item,name);
	      buttonCount++;
	      if (buttonCount==2) {
		group=new ButtonGroup();
		group.add(lastButton);
	      }
	      if (buttonCount>1)
		group.add(item);
	      lastButton=item;
	    }
	    else if (isJLabel(typeName)) {
	      JLabel label=getJLabel(name);
	      set(var,label);

	      tryToResolve(menuList,itemList,label,name);
	      buttonCount=0;
	    }
	    else if (isJSeparator(typeName)) {
	      JSeparator item=new JSeparator();
	      set(var,item);

	      tryToResolve(menuList,itemList,item,name);
	      buttonCount=0;
	    }
	    else if (isFiller(typeName)) {
	      Box.Filler filler=getBoxFiller(name);
	      set(var,filler);

	      hasResolved(menuList,filler,pathName(name,'_',false));
	      buttonCount=0;
	    }
	    else if (isJMenuBar(typeName)) {
	      JMenuBar menuBar=getJMenuBar();
	      set(var,menuBar);

	      menuList.addElement(menuBar);
	      menuList.addElement(name);
	      buttonCount=0;
	    }
	    else if (isJToolBar(typeName)) {
	      JToolBar toolBar=getJToolBar(name);
	      set(var,toolBar);

	      menuList.addElement(toolBar);
	      menuList.addElement(name);
	      buttonCount=0;
	    }
	    else
	      buttonCount=0;
	  }
	  catch(Exception cannotAccess) {
	    //System.out.println(cannotAccess);
	    //cannotAccess.printStackTrace();
	  }
	}
      }
      
      // resolve the hierarchy and build the menu, if any
      if (menuList.size()>0) {
	int size=itemList.size()/2;
	for(int i=0;i<size;i++)
	  hasResolved(menuList,
		      (JComponent)itemList.elementAt(2*i),
		      (String)itemList.elementAt(2*i+1));
      }
    }
  }

  /**
   * try to resolve the menu hierarchy, store if fails
   */
  void tryToResolve(Vector menuList,Vector storeList,JComponent item,String name) {
    String parent=pathName(name,'_',false);

    if (parent.length()>0 && !hasResolved(menuList,item,parent)) {
      storeList.addElement(item);
      storeList.addElement(parent);
      //System.out.println("store "+name+" for parent "+parent);
    }
  }

  /**
   * try to resolve the menu hierarchy
   */
  boolean hasResolved(Vector menuList,Component item,String parent) {
    int size=menuList.size()/2;

    for(int i=0;i<size;i++) {
      //System.out.println(parent+" =?= "+menuList.elementAt(2*i+1));
      if (parent.equals((String)menuList.elementAt(2*i+1))) {
	((JComponent)menuList.elementAt(2*i)).add(item);
	return true;
      }
    }
    return false;
  }

  /**
   * tests on Class: convenient methods
   */
  public static boolean isString(Class c) {
    return isString(c.getName());
  }
  public static boolean isString(String className) {
    return className.equals("java.lang.String");
  }
  public static boolean isIcon(Class c) {
    return isIcon(c.getName());
  }
  public static boolean isImageIcon(Class c) {
    return isImageIcon(c.getName());
  }
  public static boolean isIcon(String className) {
    return className.equals("javax.swing.Icon");
  }
  public static boolean isImageIcon(String className) {
    return className.equals("javax.swing.ImageIcon");
  }
  public static boolean isJButton(Class c) {
    return isJButton(c.getName());
  }
  public static boolean isJButton(String className) {
    return className.equals("javax.swing.JButton");
  }
  public static boolean isJLabel(Class c) {
    return isJLabel(c.getName());
  }
  public static boolean isJLabel(String className) {
    return className.equals("javax.swing.JLabel");
  }
  public static boolean isJSeparator(Class c) {
    return isJSeparator(c.getName());
  }
  public static boolean isJSeparator(String className) {
    return className.equals("javax.swing.JSeparator");
  }
  public static boolean isFiller(Class c) {
    return isFiller(c.getName());
  }
  public static boolean isFiller(String className) {
    return className.equals("javax.swing.Box$Filler");
  }
  public static boolean isJToggleButton(Class c) {
    return isJToggleButton(c.getName());
  }
  public static boolean isJToggleButton(String className) {
    return className.equals("javax.swing.JToggleButton");
  }
  public static boolean isJMenu(Class c) {
    return isJMenu(c.getName());
  }
  public static boolean isJMenu(String className) {
    return className.equals("javax.swing.JMenu");
  }
  public static boolean isJMenuBar(Class c) {
    return isJMenuBar(c.getName());
  }
  public static boolean isJMenuBar(String className) {
    return className.equals("javax.swing.JMenuBar");
  }
  public static boolean isJToolBar(Class c) {
    return isJToolBar(c.getName());
  }
  public static boolean isJToolBar(String className) {
    return className.equals("javax.swing.JToolBar");
  }
  public static boolean isJMenuItem(Class c) {
    return isJMenuItem(c.getName());
  }
  public static boolean isJMenuItem(String className) {
    return className.equals("javax.swing.JMenuItem");
  }
  public static boolean isJRadioButtonMenuItem(Class c) {
    return isJRadioButtonMenuItem(c.getName());
  }
  public static boolean isJRadioButtonMenuItem(String className) {
    return className.equals("javax.swing.JRadioButtonMenuItem");
  }
  public static boolean isJCheckBoxMenuItem(Class c) {
    return isJCheckBoxMenuItem(c.getName());
  }
  public static boolean isJCheckBoxMenuItem(String className) {
    return className.equals("javax.swing.JCheckBoxMenuItem");
  }
  public static boolean isJPopupMenu(Class c) {
    return isJPopupMenu(c.getName());
  }
  public static boolean isJPopupMenu(String className) {
    return className.equals("javax.swing.JPopupMenu");
  }
  public static boolean isToggleButtonMenuItem(Class c) {
    return isToggleButtonMenuItem(c.getName());
  }
  public static boolean isToggleButtonMenuItem(String className) {
    return className.equals("adavid.swing.ToggleButtonMenuItem");
  }
  public static boolean isRadioButtonMenuItem(Class c) {
    return isRadioButtonMenuItem(c.getName());
  }
  public static boolean isRadioButtonMenuItem(String className) {
    return className.equals("adavid.swing.RadioButtonMenuItem");
  }
  public static boolean isJRadioButton(Class c) {
    return isJRadioButton(c.getName());
  }
  public static boolean isJRadioButton(String className) {
    return className.equals("javax.swing.JRadioButton");
  }
  public static boolean isToggleCheckButtonMenuItem(Class c) {
    return isToggleCheckButtonMenuItem(c.getName());
  }
  public static boolean isToggleCheckButtonMenuItem(String className) {
    return className.equals("adavid.swing.ToggleCheckButtonMenuItem");
  }
  public static boolean isButtonMenuItem(Class c) {
    return isButtonMenuItem(c.getName());
  }
  public static boolean isButtonMenuItem(String className) {
    return className.equals("adavid.swing.ButtonMenuItem");
  }
  public static boolean isColor(String className) {
    return className.equals("java.awt.Color");
  }


  public static void setTitleLabel(JLabel label) {
    titleLabel=label;
  }

  public static void setProgressBar(JProgressBar progress) {
    progressBar=progress;
  }

}



