/**
 * 
 * ResourceLoader
 * useful functions to load resources
 * This class is able to auto-initialize
 * resources of another class by using reflection.
 * This is useful if one wants to change the
 * language of an application, just change
 * a configuration file and that's all, no
 * recompilation or whatever.
 * @version 1.1
 * @author Alexandre David
 */

/**
 * version history
 * 1.0: first version
 * 1.1 990829: use swing 1.1.1
 */

package adavid.util;

import java.util.*;
import javax.swing.*;
import java.net.*;
import java.io.*;
import adavid.awt.*;
import java.awt.event.*;
import java.awt.*;
import java.lang.reflect.Field;
import adavid.swing.*;

public class ResourceLoader extends Bundle {

  /**
   * extensions of the resource names to define
   * attributes
   */
  public static final String TOOLTIP_DEF=".tooltip";
  public static final String DISABLED_DEF=".disabled";
  public static final String DISABLED_SELECTED_DEF=".disabled.selected";
  public static final String PRESSED_DEF=".pressed";
  public static final String ROLLOVER_DEF=".rollover";
  public static final String ROLLOVER_SELECTED_DEF=".rollover.selected";
  public static final String SELECTED_DEF=".selected";
  public static final String ROLLOVER_ENABLED_DEF=".rollover.enabled";
  public static final String ENABLED_DEF=".enabled";
  public static final String BORDER_PAINTED_DEF=".border.painted";
  public static final String FOCUS_BORDER_DEF=".border.focused";
  public static final String FOCUS_PAINTED=".focus.painted";
  public static final String RAISED_BORDER_DEF=".border.raised";
  public static final String HIDE_NAME_DEF=".hidename";
  public static final String GLUE_NAME="GLUE";
  public static final String HORIZONTAL_GLUE_NAME="HGLUE";
  public static final String VERTICAL_GLUE_NAME="VGLUE";
  public static final char H_STRUT='H',V_STRUT='V',AREA='A';
  public static final String FLOATABLE_DEF=".floatable";
  public static final String EXTENDED_DEF="options.extended";

  public static int textPosition=JButton.RIGHT;

  boolean extendedOptionsEnabled=false;
  ActionListener listener;

  /**
   * constructor with a class
   */
  public ResourceLoader(Class target) {
    super(target);
  }

  /**
   * constructor: takes the name of the resource bundle, and
   * tries to locate the resource bundle
   */
  public ResourceLoader(String bundleName,Class target) {
    super(bundleName,target);
  }

  /**
   * constructor: takes the resource bundle in argument
   */
  public ResourceLoader(ResourceBundle bundle) {
    super(bundle);
  }

  /**
   * change the ActionListener
   */
  public void setActionListener(ActionListener listener) {
    this.listener=listener;
  }


  /**
   * get an icon by its resource name (without .icon extension)
   * return null if not found
   * the default name is resourceName.gif
   * where the '.' in the resourceName have been replaced by '_'
   */
  public ImageIcon getIcon(String def) {
    return loadIcon(getResource(def+ICON_DEF,iconName(def)));
  }



  /**
   * load an icon, given its name
   * the icon is path+name
   * return the icon, or null if the name is null or the icon was not found
   */
  public ImageIcon loadIcon(String name) {
    Image image=loadImage(name);
    if (image!=null)
      return new ImageIcon(image);
    else
      return null;
  }

  /**
   * convenient function: load a tooltip definition
   */
  public String getToolTipResource(String def,String defaultValue) {
    return getResource(def+TOOLTIP_DEF,defaultValue);
  }

  /**
   * initialize an AbstractButton item
   * the name is the resource definition name
   */
  public AbstractButton setupItem(AbstractButton item,String name) {
    item.setActionCommand(name);
    if (extendedOptionsEnabled || !(item instanceof JMenu))
      item.setIcon(getIcon(name));
    if (item.getText().length()==0) {
      if (item.getIcon()==null)
	item.setText(name);
      else
	item.setMargin(new Insets(0,0,0,0));
    } else if ((item instanceof JButton || item instanceof JToggleButton) //hide the text in this case
	       && item.getIcon()!=null && getBoolResource(name+HIDE_NAME_DEF,TRUE)) {
      item.setText("");
      item.setMargin(new Insets(0,0,0,0));
    }
    
    boolean extendedOpts=(item instanceof JButton || item instanceof JToggleButton || item instanceof JRadioButton);
    if (item.getIcon()!=null) { //to have different icons, the basic normal one is required!
      if (extendedOptionsEnabled) {
	item.setDisabledIcon(getIcon(name+DISABLED_DEF));
	if (extendedOpts || item instanceof ToggleButtonMenuItem)
	  item.setDisabledSelectedIcon(getIcon(name+DISABLED_SELECTED_DEF));
      }
      if (extendedOpts && extendedOptionsEnabled) {
	item.setPressedIcon(getIcon(name+PRESSED_DEF));
	item.setRolloverSelectedIcon(getIcon(name+ROLLOVER_SELECTED_DEF));
	item.setFocusPainted(getBoolResource(name+FOCUS_PAINTED,TRUE));
	item.setRolloverEnabled(getBoolResource(name+ROLLOVER_ENABLED_DEF,FALSE));
	item.setRolloverIcon(getIcon(name+ROLLOVER_DEF));
      }
      if (extendedOptionsEnabled
	  || item instanceof ToggleButtonMenuItem
	  || item instanceof JToggleButton) {
	item.setSelectedIcon(getIcon(name+SELECTED_DEF));

	//improve configuration
	Icon icon=item.getPressedIcon();
	if (icon==null)
	  icon=item.getRolloverIcon();
	if (icon==null)
	  icon=item.getSelectedIcon();
	if (icon!=null) {
	  item.setPressedIcon(icon);
	  if (item.getSelectedIcon()==null)
	    item.setSelectedIcon(icon);
	}
      }
    }
    if (extendedOpts) {
      String initBool=(item instanceof JRadioButton)?FALSE:TRUE;
      if (getBoolResource(name+FOCUS_BORDER_DEF,initBool))
	FocusableBorder.add(item);
      else if (extendedOptionsEnabled)
	item.setBorderPainted(getBoolResource(name+BORDER_PAINTED_DEF,FALSE));

      if (extendedOptionsEnabled) {
	if (getBoolResource(name+RAISED_BORDER_DEF,FALSE))
	  RaiseableBorder.add(item);
	item.setSelected(getBoolResource(name+SELECTED_DEF,FALSE));
      }
    }
    item.setHorizontalTextPosition(textPosition);

    if (!(item instanceof JMenu)) {
      if (actionHandler!=null) {
	try {
	  ActionManager.add(item,shortName(name,'_'),actionHandler);
	  item.addActionListener(ActionManager.listener);
	}
	catch(Exception whatever){
	  if (listener!=null) //no ActionListener for a JMenu!
	    item.addActionListener(listener);
	}
      }
      else if (listener!=null)
	item.addActionListener(listener);
    }

    if (extendedOptionsEnabled) {
      item.setEnabled(getBoolResource(name+ENABLED_DEF,TRUE));
    
      String mnemoString=getResource(name+MNEMO_DEF,"");
      if (mnemoString.length()>0) {
	char mnemo=mnemoString.charAt(0);
	item.setMnemonic(mnemo);
	if (item instanceof JMenuItem && !(item instanceof JMenu))
	  ((JMenuItem)item).setAccelerator(KeyStroke.getKeyStroke(Character.toUpperCase(mnemo),
								  ActionEvent.ALT_MASK));
      }
    }
    return item;
  }

  /**
   * returns a preset progress bar
   */
  public static JProgressBar getProgressBar() {
    JProgressBar progress=new JProgressBar();
    progress.setBorder(BorderFactory.createLoweredBevelBorder());
    progress.setBorderPainted(true);
    progress.setForeground(Color.blue);
    progress.setMinimum(0);
    progress.setMaximum(100);
    progress.setValue(0);
    return progress;
  }

  /**
   * setup a button
   */
  public AbstractButton setupButton(AbstractButton button,String name,String def) {
    button.setToolTipText(getToolTipResource(name,def));
    return setupItem(button,name);
  }

  /**
   * returns the corresponding button resource with the name definition
   */
  public JButton getJButton(String name) {
    String def=shortName(name,'_');
    JButton button=new JButton(getNameResource(name,def));
    return (JButton)setupButton(button,name,def);
  }

  /**
   * returns the corresponding button resource with the name definition
   */
  public JToggleButton getJToggleButton(String name) {
    String def=shortName(name,'_');
    JToggleButton button=new JToggleButton(getNameResource(name,def));
    return (JToggleButton)setupButton(button,name,def);
  }

  /**
   * returns the corresponding button resource with the name definition
   */
  public JRadioButton getJRadioButton(String name) {
    String def=shortName(name,'_');
    JRadioButton button=new JRadioButton(getNameResource(name,def));
    return (JRadioButton)setupButton(button,name,def);
  }

  /**
   * returns the corresponding menu
   */
  public JMenu getJMenu(String name) {
    JMenu menu=new JMenu(getNameResource(name,shortName(name,'_')));
    return (JMenu)setupItem(menu,name);
  }

  /**
   * returns the corresponding menu item
   */
  public JMenuItem getJMenuItem(String name) {
    JMenuItem item=new JMenuItem(getNameResource(name,shortName(name,'_')));
    return (JMenuItem)setupItem(item,name);
  }

  /**
   * return the corresponding popup menu
   */
  public JPopupMenu getJPopupMenu(String name) {
    JPopupMenu menu=new JPopupMenu(getNameResource(name,shortName(name,'_')));
    if (extendedOptionsEnabled)
      menu.setBorderPainted(getBoolResource(name+BORDER_PAINTED_DEF,TRUE));
    return menu;
  }

  public ToggleButtonMenuItem getToggleButtonMenuItem(String name,JToggleButton button) {
    String def=shortName(name,'_');
    ToggleButtonMenuItem item=new ToggleButtonMenuItem(getNameResource(name,def),button);
    return (ToggleButtonMenuItem)setupItem(item,name);
  }

  public ButtonMenuItem getButtonMenuItem(String name,JButton button) {
    String def=shortName(name,'_');
    ButtonMenuItem item=new ButtonMenuItem(getNameResource(name,def),button);
    return (ButtonMenuItem)setupItem(item,name);
  }

  public RadioButtonMenuItem getRadioButtonMenuItem(String name,JRadioButton button) {
    String def=shortName(name,'_');
    RadioButtonMenuItem item=new RadioButtonMenuItem(getNameResource(name,def),button);
    return (RadioButtonMenuItem)setupItem(item,name);
  }

  public JRadioButtonMenuItem getJRadioButtonMenuItem(String name) {
    JRadioButtonMenuItem item=new JRadioButtonMenuItem(getNameResource(name,shortName(name,'_')));
    return (JRadioButtonMenuItem)setupItem(item,name);
  }

  public JCheckBoxMenuItem getJCheckBoxMenuItem(String name) {
    JCheckBoxMenuItem item=new JCheckBoxMenuItem(getNameResource(name,shortName(name,'_')));
    item.setState(getBoolResource(name+STATE_DEF,FALSE));
    return (JCheckBoxMenuItem)setupItem(item,name);
  }

  public JLabel getJLabel(String name) {
    String def=shortName(name,'_');
    JLabel label=new JLabel(getNameResource(name,def));
    label.setToolTipText(getToolTipResource(name,def));
    label.setIcon(getIcon(name));
    return label;
  }

  public static Box.Filler getBoxFiller(String name) {
    Box.Filler filler=null;
    int idx=name.indexOf('$');
    if (idx<0)
      idx=name.length();
    String shortName=shortName(name.substring(0,idx),'_').toUpperCase();
    if (shortName.equals(GLUE_NAME))
      filler=(Box.Filler)Box.createGlue();
    else if (shortName.equals(HORIZONTAL_GLUE_NAME))
      filler=(Box.Filler)Box.createHorizontalGlue();
    else if (shortName.equals(VERTICAL_GLUE_NAME))
      filler=(Box.Filler)Box.createVerticalGlue();
    else {
      try {
	char c=shortName.charAt(0);
	if (c==H_STRUT)
	  filler=(Box.Filler)Box.createHorizontalStrut(Integer.parseInt(shortName.substring(1)));
	else if (c==V_STRUT)
	  filler=(Box.Filler)Box.createVerticalStrut(Integer.parseInt(shortName.substring(1)));
	else if (c==AREA) {
	  int index=shortName.indexOf('X');
	  filler=(Box.Filler)Box.createRigidArea(new Dimension(Integer.parseInt(shortName.substring(1,index)),
							       Integer.parseInt(shortName.substring(index+1))));
	}
      }
      catch(Exception whatever){}
    }
    return (filler==null)?(Box.Filler)Box.createRigidArea(new Dimension(5,5)):filler;
  }

  public static JMenuBar getJMenuBar() {
    JMenuBar menuBar=new JMenuBar();
    menuBar.setBorder(BorderFactory.createRaisedBevelBorder());
    return menuBar;
  }

  public JToolBar getJToolBar(String name) {
    JToolBar toolBar=new JToolBar();
    if (extendedOptionsEnabled) {
      toolBar.setFloatable(getBoolResource(name+FLOATABLE_DEF,TRUE));
      toolBar.setBorderPainted(getBoolResource(name+BORDER_PAINTED_DEF,TRUE));
    }
    return toolBar;
  }

}
