/**
 * StatSource
 *  * gives some statistics on source files
 * the extension of the source files may be
 * given in argument: .c c .cc cc .java java
 * the source files are read and the comments
 * are taken into account
 *
 * comments are supposed to be given as:
 * //comment => comment line, code... //comment =>code line
 * embeded comments like this one with start as first token and end as last token
 * @version 1.0
 * @author Alexandre David
 */

/**
 * version history:
 * 1.0: first version
 */

package adavid.util;

import java.io.*;
import java.util.StringTokenizer;


public class StatSource {

  static String ext=".java";

    /**
     * information on a file
     */
  static SourceInfo infoFile(File source) {
    SourceInfo info=new SourceInfo(source);
    info.nbClass=1;
    try {
      BufferedReader reader=new BufferedReader(new FileReader(source));
      String line;
      boolean comment=false;
      while((line=reader.readLine())!=null) {
	StringTokenizer tokenizer=new StringTokenizer(line);
	if (tokenizer.countTokens()==0) {
	  //System.out.println("    : "+line);
	  info.emptyLines++;
	}
	else {
	  String first=tokenizer.nextToken();
	  String last=first;
	  while(tokenizer.hasMoreTokens())
	    last=tokenizer.nextToken();

	  if (comment) {
	    info.commentLines++;
	    //System.out.println("****: "+line);
	    if (last.indexOf("*/")>=0)
	      comment=false;
	    //if there is something else on this line, we do not care
	  }
	  else {
	    if (first.indexOf("//")>=0) {
	      //System.out.println("****: "+line);
	      info.commentLines++;
	    }
	    else if (first.indexOf("/*")>=0) {
	      info.commentLines++;
	      //System.out.println("****: "+line);
	      comment=true;
	    }
	    else if (last==first && first.equals("}")) {
	      //System.out.println("  {:"+line);
	      info.falseCode++;
	    }
	    else {
	      //System.out.println("code: "+line);
	      info.codeLines++;
	    }
	  }
	}
      }
      reader.close();
    }
    catch(Exception whatever) {
      System.err.println(whatever);
    }
    System.out.println(info);
    return info;
  }

    /**
     * information on a directory
     */
  static SourceInfo infoDir(File dir) {
    SourceInfo info=new SourceInfo(dir);
    try {
      String[] list=dir.list();
      for (int i=0;i<list.length;i++) {
	File file=new File(dir,list[i]);
	if (file.isDirectory())
	  info.add(infoDir(file));
	else if (JMakefile.isExtFile(list[i],ext))
	  info.add(infoFile(file));
      }
    }
    catch(Exception whatever) {
      System.err.println(whatever);
    }
    System.out.println(info);
    return info;   
  }

  public static void main(String[] args) {
    if (args.length>0) {
      if (args[0].charAt(0)=='.')
	ext=args[0];
      else
	ext='.'+args[0];
    }
    infoDir(new File("."));
  }
}

/**
 * this class contains information on a source file
 */
class SourceInfo {
  File source;
  int nbClass=0;
  int commentLines=0;
  int codeLines=0;
  int emptyLines=0;
  int falseCode=0;
  long len=0;

  public SourceInfo(File file) {
    source=file;
    if (!file.isDirectory())
      len=file.length();
  }

  public String toString() {
    StringBuffer buffer=new StringBuffer(128);
    //the thing with the buffer is done manually since it
    //is done approximately this way by the compiler for "..."+"blabla"
    buffer.append(source.toString());
    if (nbClass>1)
      buffer.append(" (").append(Integer.toString(nbClass)).append(" files)");
    buffer.append(" : code=").append(codeLines).append(" comments=").append(commentLines).append(" single\"}\"=").append(falseCode).append(" empty=").append(emptyLines).append(" : total=").append(codeLines+commentLines+falseCode+emptyLines).append(" # size=").append(len);
    return buffer.toString();
  }

    /**
     * concat file information
     */
  public void add(SourceInfo info) {
    nbClass+=info.nbClass;
    commentLines+=info.commentLines;
    codeLines+=info.codeLines;
    emptyLines+=info.emptyLines;
    len+=info.len;
    falseCode+=info.falseCode;
  }

}




